# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines plotting classes for the data series defined in
`~gwpy.data`
"""

import copy
import numpy

from matplotlib.projections import register_projection
from matplotlib import (cm, colors)

from . import (tex, rcParams)
from .utils import *
from .core import Plot
from .axes import Axes
from .decorators import auto_refresh
from ..spectrum import (Spectrum, SpectralVariance)

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class SpectrumAxes(Axes):
    """Custom `Axes` for a :class:`~gwpy.plotter.SpectrumPlot`.
    """
    name = 'spectrum'

    # -------------------------------------------
    # GWpy class plotting methods

    @auto_refresh
    def plot(self, *args, **kwargs):
        """Plot data onto these Axes.

        Parameters
        ----------
        args
            a single :class:`~gwpy.spectrum.core.spectrum` (or sub-class)
            or standard (x, y) data arrays
        kwargs
            keyword arguments applicable to :meth:`~matplotib.axes.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        if len(args) == 1 and isinstance(args[0], Spectrum):
            return self.plot_spectrum(*args, **kwargs)
        elif len(args) == 1 and isinstance(args[0], SpectralVariance):
            return self.plot_variance(*args, **kwargs)
        else:
            return super(SpectrumAxes, self).plot(*args, **kwargs)

    @auto_refresh
    def plot_spectrum(self, spectrum, **kwargs):
        """Plot a :class:`~gwpy.spectrum.core.Spectrum` onto these axes

        Parameters
        ----------
        spectrum : :class:`~gwpy.spectrum.core.Spectrum`
            data to plot
        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        if tex.USE_TEX:
            kwargs.setdefault('label', tex.label_to_latex(spectrum.name))
        else:
            kwargs.setdefault('label', spectrum.name)
        if not kwargs.get('label', True):
            kwargs.pop('label')
        line = self.plot(spectrum.frequencies.value, spectrum.value, **kwargs)
        if len(self.lines) == 1:
            try:
                self.set_xlim(spectrum.frequencies[0].value,
                              spectrum.frequencies[-1].value +
                              spectrum.df.value)
            except IndexError:
                pass
        if 'label' in kwargs:
            self.legend()
        return line

    @auto_refresh
    def plot_spectrum_mmm(self, mean_, min_=None, max_=None, alpha=0.1,
                          **kwargs):
        """Plot a `Spectrum` onto these axes, with (min, max) shaded
        regions

        The `mean_` `Spectrum` is plotted normally, while the `min_`
        and `max_ spectra are plotted lightly below and above,
        with a fill between them and the mean_.

        Parameters
        ----------
        mean_ : :class:`~gwpy.spectrum.core.Spectrum
            data to plot normally
        min_ : :class:`~gwpy.spectrum.core.Spectrum
            first data set to shade to mean_
        max_ : :class:`~gwpy.spectrum.core.Spectrum
            second data set to shade to mean_
        alpha : `float`, optional
            weight of filled region, ``0.0`` for transparent through ``1.0``
            opaque
        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`

        Returns
        -------
        artists : `tuple`
            a 5-tuple containing (Line2d for mean_, `Line2D` for min_,
            `PolyCollection` for min_ shading, `Line2D` for max_, and
            `PolyCollection` for max_ shading)

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        # plot mean
        line1 = self.plot_spectrum(mean_, **kwargs)[0]
        # plot min and max
        kwargs.pop('label', None)
        color = kwargs.pop('color', line1.get_color())
        linewidth = kwargs.pop('linewidth', line1.get_linewidth()) / 10
        if min_ is not None:
            a = self.plot(min_.frequencies.value, min_.value, color=color,
                          linewidth=linewidth, **kwargs)
            if alpha:
                b = self.fill_between(min_.frequencies.value, mean_.value,
                                      min_.value, alpha=alpha, color=color)
            else:
                b = None
        else:
            a = b = None
        if max_ is not None:
            c = self.plot(max_.frequencies.value, max_.value, color=color,
                          linewidth=linewidth, **kwargs)
            if alpha:
                d = self.fill_between(max_.frequencies.value, mean_.value,
                                      max_.value, alpha=alpha, color=color)
            else:
                d = None
        else:
            c = d = None
        return line1, a, b, c, d

    @auto_refresh
    def plot_variance(self, specvar, norm='log', **kwargs):
        """Plot a :class:`~gwpy.spectrum.hist.SpectralVariance` onto
        these axes

        Parameters
        ----------
        spectrum : class:`~gwpy.spectrum.hist.SpectralVariance`
            data to plot
        **kwargs
            any other eyword arguments acceptable for
            :meth:`~matplotlib.Axes.pcolormesh`

        Returns
        -------
        MeshGrid
            the :class:`~matplotlib.collections.MeshGridD` for this layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.pcolormesh`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        if norm == 'log':
            vmin = kwargs.pop('vmin', None)
            vmax = kwargs.pop('vmax', None)
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        kwargs['norm'] = norm
        x = numpy.concatenate((specvar.frequencies.value,
                               [specvar.x0.value +
                                specvar.dx.value * specvar.shape[0]]))
        y = specvar.bins.value
        X, Y = numpy.meshgrid(x, y, copy=False, sparse=True)
        mesh = self.pcolormesh(X, Y, specvar.value.T, **kwargs)
        if len(self.collections) == 1:
            self.set_yscale('log', nonposy='mask')
            self.set_xlim(x[0], x[-1])
            self.set_ylim(y[0], y[-1])
        return mesh


register_projection(SpectrumAxes)


class SpectrumPlot(Plot):
    """`Figure` for displaying a :class:`~gwpy.spectrum.core.Spectrum`.
    """
    _DefaultAxesClass = SpectrumAxes

    def __init__(self, *series, **kwargs):
        kwargs.setdefault('projection', self._DefaultAxesClass.name)
        # extract custom keyword arguments
        sep = kwargs.pop('sep', False)
        xscale = kwargs.pop(
            'xscale', kwargs.pop('logx', True) and 'log' or 'linear')
        yscale = kwargs.pop(
            'yscale', kwargs.pop('logy', True) and 'log' or 'linear')
        sharex = kwargs.pop('sharex', False)
        sharey = kwargs.pop('sharey', False)
        # separate custom keyword arguments
        axargs, plotargs = self._parse_kwargs(kwargs)

        # initialise figure
        super(SpectrumPlot, self).__init__(**kwargs)

        # plot data
        x0 = []
        axesdata = self._get_axes_data(series, sep=sep)
        for data in axesdata:
            ax = self._add_new_axes(**axargs)
            for fs in data:
                ax.plot(fs, **plotargs)
            x0.append(min([fs.df.value for fs in data]))
            if 'sharex' not in axargs and sharex is True:
                axargs['sharex'] = ax
            if 'sharey' not in axargs and sharey is True:
                axargs['sharey'] = ax
        if sharex:
            x0 = [min(x0)]*len(x0)
        axargs.pop('sharex', None)
        axargs.pop('sharey', None)
        axargs.pop('projection', None)

        for i, ax in enumerate(self.axes):
            # format axes
            for key, val in axargs.iteritems():
                getattr(ax, 'set_%s' % key)(val)
            # fix log frequency scale with f0 = DC
            if xscale in ['log']:
                xlim = list(ax.get_xlim())
                if not xlim[0]:
                    xlim[0] = x0[i]
                ax.set_xlim(*xlim)
            # set axis scales
            ax.set_xscale(xscale)
            ax.set_yscale(yscale)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013-2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting a Rayleigh-statistic `Spectum`

"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.spectrum'

# First, we import the :class:`~gwpy.timeseries.TimeSeries` and :meth:`~gwpy.timeseries.TimeSeries.get` the data:
from gwpy.timeseries import TimeSeries
gwdata = TimeSeries.get(
    'H1:LDAS-STRAIN,rds', 'September 16 2010 06:40', 'September 16 2010 06:50')

# Next, we can calculate a Rayleigh statistic `Spectrum` using the 
# :meth:`~gwpy.timeseries.TimeSeries.rayleigh_spectrum` method of the
# `~gwpy.timeseries.TimeSeries` with a 2-second FFT and 1-second overlap (50%):
rayleigh = gwdata.rayleigh_spectrum(2, 1)

# and can make a plot using the :meth:`~Spectrum.plot` method
plot = rayleigh.plot()
plot.set_xscale('log')
plot.set_xlim(40, 4000)
plot.set_xlabel('Frequency [Hz]')
plot.set_ylabel('Rayleigh statistic')
plot.show()



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>"""GWpy Example: plotting a time-series

"""Plotting a filter

I would like to look at the Bode representation of a linear filter.
"""

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__currentmodule__ = 'gwpy.plotter'

# First, we import the objects we need
from math import pi
from scipy import signal
from gwpy.plotter import BodePlot

# Now, we can calculate a fourth-order butterworth filter using the
# :meth:`~scipy.signal.butter` function
highpass = signal.butter(4, 10 * (2. * pi), btype='highpass',
                         analog=True)

# The `BodePlot` knows how to plot filters:
plot = BodePlot(highpass)
plot.maxes.set_title('10\,Hz high-pass filter (4th order Butterworth)')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from an Omega-format ASCII file.
"""

import sys
from math import sqrt

from numpy import loadtxt

from astropy.io import registry

from .ascii import table_from_ascii_factory
from ..lsctables import (SnglBurstTable, SnglBurst)
from ... import version
from ...io.cache import file_list
from ...time import LIGOTimeGPS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

OMEGA_COLUMNS = ['search', 'event_id', 'peak_time', 'peak_time_ns',
                 'start_time', 'start_time_ns',
                 'stop_time', 'stop_time_ns', 'duration',
                 'central_freq', 'flow', 'fhigh', 'bandwidth',
                 'snr', 'amplitude', 'confidence']

OMEGADQ_COLUMNS = ['search', 'event_id', 'peak_time', 'peak_time_ns',
                   'start_time', 'start_time_ns',
                   'stop_time', 'stop_time_ns', 'duration',
                   'ms_start_time', 'ms_start_time_ns',
                   'ms_stop_time', 'ms_stop_time_ns', 'ms_duration',
                   'central_freq', 'peak_frequency',
                   'flow', 'fhigh', 'bandwidth',
                   'ms_flow', 'ms_fhigh', 'ms_bandwidth',
                   'snr', 'ms_snr', 'amplitude', 'confidence']


def sngl_burst_from_omega(line, columns=OMEGA_COLUMNS):
    """Build a `SnglBurst` from an Omega ASCII line.

    Parameters
    ----------
    line : `str`
        the ASCII line to read
    columns : `list`
        a `list` of valid `LIGO_LW` column names to load.

    Returns
    -------
    burst : :class:`~gwpy.table.lsctables.SnglBurst`
        a `SnglBurst` with attributes populated from the ROOT data.
    """
    t = SnglBurst()
    t.search = u'omega'

    if isinstance(line, str):
        line = line.rstrip('\n').split()
    line = map(float, line)
    try:
        peak, freq, duration, band, nerg, clst_size, clst_nerg, clst_N = line
    except IndexError:
        peak, freq, duration, band, nerg = line

    # parse time data
    peak = LIGOTimeGPS(peak)
    if 'time' in columns or 'peak_time' in columns:
        t.peak_time = peak.seconds
    if 'time' in columns or 'peak_time_ns' in columns:
        t.peak_time_ns = peak.nanoseconds
    start = peak - duration / 2.
    if 'start_time' in columns:
        t.start_time = start.seconds
    if 'start_time_ns' in columns:
        t.start_time_ns = start.nanoseconds
    stop = peak + duration / 2.
    if 'stop_time' in columns:
        t.stop_time = stop.seconds
    if 'stop_time_ns' in columns:
        t.stop_time_ns = stop.nanoseconds
    if 'duration' in columns:
        t.duration = duration

    # parse frequency data
    flow = freq - band / 2.
    fhigh = freq + band / 2.
    if 'central_freq' in columns:
        t.central_freq = freq
    if 'flow' in columns:
        t.flow = flow
    if 'fhigh' in columns:
        t.fhigh = fhigh
    if 'bandwidth' in columns:
        t.bandwidth = band

    # others
    snr = sqrt(2 * nerg)
    if 'snr' in columns:
        t.snr = snr
    if 'amplitude' in columns:
        t.amplitude = nerg
    if 'confidence' in columns:
        t.confidence = snr

    return t


def sngl_burst_from_omegadq(line, columns=OMEGADQ_COLUMNS):
    """Build a `SnglBurst` from an Omega DQ (DetChar) ASCII line.

    Parameters
    ----------
    line : `str`
        the ASCII line to read
    columns : `list`
        a `list` of valid `LIGO_LW` column names to load.

    Returns
    -------
    burst : :class:`~gwpy.table.lsctables.SnglBurst`
        a `SnglBurst` with attributes populated from the ROOT data.
    """
    t = SnglBurst()
    t.search = u'omega'

    if isinstance(line, str):
        line = map(float, line.rstrip('\n').split())
    (start, stop, peak, flow, fhigh, nevents, ms_start, ms_stop, ms_flow,
     ms_fhigh, clst_size, clst_energy, ms_snr) = line

    # parse time data
    peak = LIGOTimeGPS(peak)
    if 'time' in columns or 'peak_time' in columns:
        t.peak_time = peak.seconds
    if 'time' in columns or 'peak_time_ns' in columns:
        t.peak_time_ns = peak.nanoseconds
    start = LIGOTimeGPS(start)
    if 'start_time' in columns:
        t.start_time = start.seconds
    if 'start_time_ns' in columns:
        t.start_time_ns = start.nanoseconds
    stop = LIGOTimeGPS(stop)
    if 'stop_time' in columns:
        t.stop_time = stop.seconds
    if 'stop_time_ns' in columns:
        t.stop_time_ns = stop.nanoseconds
    if 'duration' in columns:
        t.duration = float(stop - start)

    # parse frequency data
    flow = flow
    fhigh = fhigh
    if 'flow' in columns:
        t.flow = flow
    if 'fhigh' in columns:
        t.fhigh = fhigh
    if 'bandwidth' in columns:
        t.bandwidth = fhigh-flow
    if 'central_freq' in columns:
        t.central_freq = flow + (fhigh - flow) * .5

    # most-significant tile information
    ms_start = LIGOTimeGPS(ms_start)
    if 'ms_start_time' in columns:
        t.ms_start_time = ms_start.seconds
    if 'ms_start_time_ns' in columns:
        t.ms_start_time_ns = ms_start.nanoseconds
    ms_stop = LIGOTimeGPS(ms_stop)
    if 'ms_stop_time' in columns:
        t.ms_stop_time = ms_stop.seconds
    if 'ms_stop_time_ns' in columns:
        t.ms_stop_time_ns = ms_stop.nanoseconds
    if 'ms_duration' in columns:
        t.ms_duration = float(ms_stop - ms_start)
    if 'ms_flow' in columns:
        t.ms_flow = ms_flow
    if 'ms_fhigh' in columns:
        t.ms_fhigh = ms_fhigh
    if 'ms_bandwidth' in columns:
        t.ms_bandwidth = ms_fhigh - ms_flow
    if 'peak_frequency' in columns:
        t.peak_frequency = ms_flow + (ms_fhigh - ms_flow) * .5

    # others
    snr = sqrt(2 * clst_energy)
    if 'snr' in columns:
        t.snr = snr
    if 'amplitude' in columns:
        t.amplitude = clst_energy
    if 'confidence' in columns:
        t.confidence = snr
    if 'ms_snr' in columns:
        t.ms_snr = sqrt(2 * ms_snr)

    return t


# register OmegaDQ
registry.register_reader(
    'omegadq', SnglBurstTable,
    table_from_ascii_factory(SnglBurstTable, 'omegadq',
                             sngl_burst_from_omegadq, OMEGADQ_COLUMNS))

# register Omega
registry.register_reader(
    'omega', SnglBurstTable,
    table_from_ascii_factory(SnglBurstTable, 'omega',
                             sngl_burst_from_omega, OMEGA_COLUMNS,
                             comments='%'))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the `DataQualityFlag` and `DataQualityDict`.

The `DataQualityFlag` represents an annotated set of data-quality segments
indicating something about the state of a laser-interferometer
gravitational-wave detector in a given time interval.

The `DataQualityDict` is just a `dict` of flags, provided as a convenience
for handling multiple flags over the same global time interval.
"""

import operator
import re
import warnings
from urlparse import urlparse
from copy import copy as shallowcopy
from math import (floor, ceil)
from threading import Thread
from Queue import Queue

from numpy import inf

from glue.segments import PosInfinity

from .. import version
from ..time import to_gps
from ..utils.deps import with_import
from ..utils.compat import OrderedDict
from ..io import (reader, writer)
from .segments import Segment, SegmentList

__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = ['DataQualityFlag', 'DataQualityDict']

re_IFO_TAG_VERSION = re.compile(r"\A(?P<ifo>[A-Z]\d):(?P<tag>[^/]+):(?P<version>\d+)\Z")
re_IFO_TAG = re.compile(r"\A(?P<ifo>[A-Z]\d):(?P<tag>[^/]+)\Z")
re_TAG_VERSION = re.compile(r"\A(?P<tag>[^/]+):(?P<ver>\d+)\Z")


class DataQualityFlag(object):
    """A representation of a named set of segments.

    Parameters
    ----------
    name : str, optional
        The name of this flag.
        This should be of the form {ifo}:{tag}:{version}, e.g.
        'H1:DMT-SCIENCE:1'. Use `label` for human-readable names.

    active : `SegmentList`, optional
        A list of active segments for this flag

    known : `SegmentList`, optional
        A list of known segments for this flag

    label : `str`, optional
        Human-readable name for this flag, e.g. ``'Science-mode'``

    category : `int`, optional
        Veto category for this flag.

    description : `str`, optional
        Human-readable description of why this flag was created.

    isgood : `bool`, optional
        Do active segments mean the IFO was in a good state?
    """
    _EntryClass = Segment
    _ListClass = SegmentList

    def __init__(self, name=None, active=None, known=None, label=None,
                 category=None, description=None, isgood=True, padding=None,
                 valid=None):
        """Define a new DataQualityFlag.
        """
        self.name = name
        if valid is not None:
            if known is not None:
                raise ValueError("Please give only 'known', and not both "
                                 "'known' and 'valid'")
            self.valid = valid
        else:
            self.known = known
        self.active = active
        self.label = label
        self.category = category
        self.description = description
        self.isgood = isgood
        self.padding = padding

    # -------------------------------------------------------------------------
    # read-write properties

    @property
    def name(self):
        """The name associated with this flag.

        This normally takes the form {ifo}:{tag}:{version}. If found,
        each component is stored separately the associated attributes.

        :type: `str`
        """
        return self._name

    @name.setter
    def name(self, n):
        self._name = n
        try:
            self._parse_name(n)
        except ValueError:
            self._parse_name(None)

    @property
    def ifo(self):
        """The interferometer associated with this flag.

        This should be a single uppercase letter and a single number,
        e.g. ``'H1'``.

        :type: `str`
        """
        return self._ifo

    @ifo.setter
    def ifo(self, ifoname):
        self._ifo = ifoname

    @property
    def tag(self):
        """The tag (name) associated with this flag.

        This should take the form ``'AAA-BBB_CCC_DDD'``, i.e. where
        each component is an uppercase acronym of alphanumeric
        characters only, e.g. ``'DCH-IMC_BAD_CALIBRATION'`` or
        ``'DMT-SCIENCE'``.

        :type: `str`
        """
        return self._tag

    @tag.setter
    def tag(self, n):
        self._tag = n

    @property
    def version(self):
        """The version number of this flag.

        Each flag in the segment database is stored with a version
        integer, with each successive version representing a more
        accurate dataset for its known segments than any previous.

        :type: `int`
        """
        return self._version

    @version.setter
    def version(self, v):
        if v is None:
            self._version = None
        else:
            self._version = int(v)

    @property
    def label(self):
        """A human-readable label for this flag.

        For example: ``'Science-mode'``.

        :type: `str`
        """
        return self._label

    @label.setter
    def label(self, lab):
        self._label = lab

    @property
    def active(self):
        """The set of segments during which this flag was
        active.
        """
        return self._active

    @active.setter
    def active(self, segmentlist):
        if segmentlist is None:
            del self.active
        else:
            self._active = self._ListClass(map(self._EntryClass, segmentlist))

    @active.deleter
    def active(self):
        self._active = self._ListClass()

    @property
    def known(self):
        """The set of segments during which this flag was
        known, and its state was well defined.
        """
        return self._known

    @known.setter
    def known(self, segmentlist):
        if segmentlist is None:
            del self.known
        else:
            self._known = self._ListClass(map(self._EntryClass, segmentlist))

    @known.deleter
    def known(self):
        self._known = self._ListClass()

    @property
    def valid(self):
        """The set of segments during which this flag was
        known, and its state was well defined.
        """
        warnings.warn("The 'valid' property of the DataQualityFlag "
                      "has been renamed 'known' and will be removed in "
                      "the near future, please move to using 'known'.",
                      DeprecationWarning)
        return self.known

    @valid.setter
    def valid(self, segmentlist):
        warnings.warn("The 'valid' property of the DataQualityFlag "
                      "has been renamed 'known' and will be removed in "
                      "the near future, please move to using 'known'.",
                      DeprecationWarning)
        self.known = segmentlist

    @valid.deleter
    def valid(self):
        warnings.warn("The 'valid' property of the DataQualityFlag "
                      "has been renamed 'known' and will be removed in "
                      "the near future, please move to using 'known'.",
                      DeprecationWarning)
        self._known = self._ListClass()

    @property
    def category(self):
        """Veto category for this flag.

        :type: `int`
        """
        return self._category

    @category.setter
    def category(self, cat):
        if cat is None:
            self._category = None
        else:
            self._category = int(cat)

    @property
    def description(self):
        """Description of why/how this flag was generated.

        :type: `str`
        """
        return self._description

    @description.setter
    def description(self, desc):
        self._description = desc

    @property
    def isgood(self):
        """Whether `active` segments mean the instrument was in a good state.

        :type: `bool`
        """
        return self._isgood

    @isgood.setter
    def isgood(self, good):
        self._isgood = bool(good)

    @property
    def padding(self):
        """[start, end) padding for this flag's active segments.
        """
        try:
            return self._padding
        except AttributeError:
            self._padding = (0, 0)
            return self.padding

    @padding.setter
    def padding(self, pad):
        if pad is None:
            self._padding = (0, 0)
        else:
            self._padding = (pad[0], pad[1])

    # -------------------------------------------------------------------------
    # read-only properties

    @property
    def texname(self):
        """Name of this flag in LaTeX printable format.
        """
        try:
            return self.name.replace('_', r'\_')
        except AttributeError:
            return None

    @property
    def extent(self):
        """The single GPS ``[start, stop)`` enclosing segment of this
        `DataQualityFlag`.

        :type: `Segment`
        """
        return self.known.extent()

    @property
    def livetime(self):
        """Amount of time this flag was `active`.

        :type: `float`
        """
        return abs(self.active)

    @property
    def regular(self):
        """`True` if the `active` segments are a proper subset of the `known`.

        :type: `bool`
        """
        return abs(self.active - self.known) == 0

    # -------------------------------------------------------------------------
    # classmethods

    @classmethod
    def query(cls, flag, *args, **kwargs):
        """Query for segments of a given flag

        This method intelligently selects the `~DataQualityFlag.query_segdb`
        or the `~DataQualityFlag.query_dqsegdb` methods based on the
        ``url`` kwarg given.

        Parameters
        ----------
        flag : str
            The name of the flag for which to query
        *args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments
        url : `str`, optional, default: ``'https://segments.ligo.org'``
            URL of the segment database

        See Also
        --------
        DataQualityFlag.query_segdb
        DataQualityFlag.query_dqsegdb
            for details on the actual query engine, and documentation of
            other keyword arguments appropriate for each query

        Returns
        -------
        flag : `DataQualityFlag`
            A new `DataQualityFlag`, with the `known` and `active` lists
            filled appropriately.
        """
        url = kwargs.get('url', 'https://segments.ligo.org')
        if 'dqsegdb' in url or re.match('https://[a-z1-9-]+.ligo.org', url):
            return cls.query_dqsegdb(flag, *args, **kwargs)
        else:
            return cls.query_segdb(flag, *args, **kwargs)

    @classmethod
    def query_segdb(cls, flag, *args, **kwargs):
        """Query the initial LIGO segment database for the given flag

        Parameters
        ----------
        flag : str
            The name of the flag for which to query
        *args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments
        url : `str`, optional, default: ``'https://segments.ligo.org'``
            URL of the segment database

        Returns
        -------
        flag : `DataQualityFlag`
            A new `DataQualityFlag`, with the `known` and `active` lists
            filled appropriately.
        """
        # parse arguments
        if len(args) == 1 and isinstance(args[0], SegmentList):
            qsegs = args[0]
        elif len(args) == 1 and len(args[0]) == 2:
            qsegs = SegmentList(Segment(map(to_gps, args[0])))
        elif len(args) == 2:
            qsegs = SegmentList([Segment(map(to_gps, args))])
        else:
            raise ValueError("DataQualityFlag.query must be called with a "
                             "flag name, and either GPS start and stop times, "
                             "or a SegmentList of query segments")
        # process query
        try:
            flags = DataQualityDict.query([flag], qsegs, **kwargs)
        except TypeError as e:
            if 'DataQualityDict' in str(e):
                raise TypeError(str(e).replace('DataQualityDict',
                                               cls.__name__))
            else:
                raise
        if len(flags) > 1:
            raise RuntimeError("Multiple flags returned for single query, "
                               "something went wrong:\n    %s"
                               % '\n    '.join(flags.keys()))
        elif len(flags) == 0:
            raise RuntimeError("No flags returned for single query, "
                               "something went wrong.")
        return flags[flag]

    @classmethod
    @with_import('dqsegdb.apicalls')
    def query_dqsegdb(cls, flag, *args, **kwargs):
        """Query the advanced LIGO DQSegDB for the given flag

        Parameters
        ----------
        flag : str
            The name of the flag for which to query
        *args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments
        url : `str`, optional, default: ``'https://segments.ligo.org'``
            URL of the segment database

        Returns
        -------
        flag : `DataQualityFlag`
            A new `DataQualityFlag`, with the `known` and `active` lists
            filled appropriately.
        """
        # parse arguments
        if len(args) == 1 and isinstance(args[0], SegmentList):
            qsegs = args[0]
        elif len(args) == 1 and len(args[0]) == 2:
            qsegs = SegmentList(Segment(map(to_gps, args[0])))
        elif len(args) == 2:
            qsegs = SegmentList([Segment(map(to_gps, args))])
        else:
            raise ValueError("DataQualityFlag.query must be called with a "
                             "flag name, and either GPS start and stop times, "
                             "or a SegmentList of query segments")
        # get server
        protocol, server = kwargs.pop(
            'url', 'https://segments.ligo.org').split('://', 1)

        # parse flag
        out = cls(name=flag)
        if out.ifo is None or out.tag is None:
            raise ValueError("Cannot parse ifo or tag (name) for flag %r"
                             % flag)

        # other keyword arguments
        request = kwargs.pop('request', 'metadata,active,known')

        # process query
        for start, end in qsegs:
            if end == PosInfinity or float(end) == +inf:
                end = to_gps('now').seconds
            if out.version is None:
                data, versions, _ = apicalls.dqsegdbCascadedQuery(
                    protocol, server, out.ifo, out.tag, request,
                    start, end)
                metadata = versions[-1]['metadata']
            else:
                data, _ = apicalls.dqsegdbQueryTimes(
                    protocol, server, out.ifo, out.tag, out.version, request,
                    start, end)
                metadata = data['metadata']
            new = cls(name=flag)
            for s2 in data['active']:
                new.active.append(Segment(*s2))
            for s2 in data['known']:
                new.known.append(Segment(*s2))
            segl = SegmentList([Segment(start, end)])
            new.known &= segl
            new.active &= segl
            out += new
            out.description = metadata.get('flag_description', None)
            out.isgood = not metadata.get(
                'active_indicates_ifo_badness', False)

        return out

    # use input/output registry to allow multi-format reading
    read = classmethod(reader(doc="""
    Read segments from file into a `DataQualityFlag`.

    Parameters
    ----------
    filename : `str`
        path of file to read
    format : `str`, optional
        source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.
    flag : `str`, optional, default: read all segments
        name of flag to read from file.
    coltype : `type`, optional, default: `float`
        datatype to force for segment times, only valid for
        ``format='segwizard'``.
    strict : `bool`, optional, default: `True`
        require segment start and stop times match printed duration,
        only valid for ``format='segwizard'``.


    Returns
    -------
    dqflag : `DataQualityFlag`
        formatted `DataQualityFlag` containing the active and known
        segments read from file.

    Notes
    -----
    When reading with ``format='segwizard'`` the
    :attr:`~DataQualityFlag.known` `SegmentList` will simply represent
    the extent of the :attr:`~DataQualityFlag.active` `SegmentList`.
    """))

    @classmethod
    def from_veto_def(cls, veto):
        """Define a `DataQualityFlag` from a `~glue.ligolw.lsctables.VetoDef`
        """
        name = '%s:%s' % (veto.ifo, veto.name)
        try:
            name += ':%d' % int(veto.version)
        except TypeError:
            pass
        if veto.end_time == 0:
            veto.end_time = PosInfinity
        known = Segment(veto.start_time, veto.end_time)
        pad = (veto.start_pad, veto.end_pad)
        return cls(name=name, known=[known], category=veto.category,
                   description=veto.comment, padding=pad)

    # -------------------------------------------------------------------------
    # instance methods

    write = writer()

    def populate(self, source='https://segments.ligo.org', segments=None,
                 pad=True, **kwargs):
        """Query the segment database for this flag's active segments.

        This method assumes all of the metadata for each flag have been
        filled. Minimally, the following attributes must be filled

        .. autosummary::

           ~DataQualityFlag.name
           ~DataQualityFlag.known

        Segments will be fetched from the database, with any
        :attr:`~DataQualityFlag.padding` added on-the-fly.

        This `DataQualityFlag` will be modified in-place.

        Parameters
        ----------
        source : `str`
            source of segments for this flag. This must be
            either a URL for a segment database or a path to a file on disk.

        segments : `SegmentList`, optional
            a list of valid segments during which to query, if not given,
            existing known segments for this flag will be used.

        pad : `bool`, optional, default: `True`
            apply the `~DataQualityFlag.padding` associated with this
            flag, default: `True`.

        **kwargs
            any other keyword arguments to be passed to
            :meth:`DataQualityFlag.query` or :meth:`DataQualityFlag.read`.

        Returns
        -------
        self : `DataQualityFlag`
            a reference to this flag
        """
        tmp = DataQualityDict()
        tmp[self.name] = self
        tmp.populate(source=source, segments=segments, pad=pad, **kwargs)
        return tmp[self.name]

    def contract(self, x):
        """Contract each of the `active` `Segments` by ``x`` seconds.

        This method adds ``x`` to each segment's lower bound, and subtracts
        ``x`` from the upper bound.

        The :attr:`~DataQualityFlag.active` `SegmentList` is modified
        in place.

        Parameters
        ----------
        x : `float`
            number of seconds by which to contract each `Segment`.
        """
        self.active = self.active.contract(x)
        return self.active

    def protract(self, x):
        """Protract each of the `active` `Segments` by ``x`` seconds.

        This method subtracts ``x`` from each segment's lower bound,
        and adds ``x`` to the upper bound, while maintaining that each
        `Segment` stays within the `known` bounds.

        The :attr:`~DataQualityFlag.active` `SegmentList` is modified
        in place.

        Parameters
        ----------
        x : `float`
            number of seconds by which to protact each `Segment`.
        """
        self.active = self.active.protract(x)
        return self.active

    def pad(self, *args):
        """Apply a padding to each segment in this `DataQualityFlag`

        This method either takes no arguments, in which case the value of
        the :attr:`~DataQualityFlag.padding` attribute will be used,
        or two values representing the padding for the start and end of
        each segment.

        For both the `start` and `end` paddings, a positive value means
        pad forward in time, so that a positive `start` pad or negative
        `end` padding will contract a segment at one or both ends,
        and vice-versa.

        This method will apply the same padding to both the
        `~DataQualityFlag.known` and `~DataQualityFlag.active` lists,
        but will not :meth:`~DataQualityFlag.coalesce` the result.

        Parameters
        ----------
        start : `float`
            padding to apply to the start of the each segment
        end : `float`
            padding to apply to the end of each segment

        Returns
        -------
        paddedflag : `DataQualityFlag`
            a view of the modified flag
        """
        if len(args) == 0:
            start, end = self.padding
        elif len(args) == 2:
            start, end = args
        else:
            raise ValueError("Cannot parse (start, end) padding from %r"
                             % args)
        new = self.copy()
        new.known = [(s[0]+start, s[1]+end) for s in self.known]
        new.active = [(s[0]+start, s[1]+end) for s in self.active]
        return new

    def round(self):
        """Round this flag to integer segments.

        Returns
        -------
        roundedflag : `DataQualityFlag`
            A copy of the original flag with the `active` and `known` segments
            padded out to the enclosing integer boundaries.
        """
        new = self.copy()
        new.active = self._ListClass([self._EntryClass(int(floor(s[0])),
                                                       int(ceil(s[1]))) for
                                     s in new.active])
        new.known = self._ListClass([self._EntryClass(int(floor(s[0])),
                                                      int(ceil(s[1]))) for
                                     s in new.known])
        return new.coalesce()

    def coalesce(self):
        """Coalesce the segments for this flag.

        This method does two things:

        - `coalesces <SegmentList.coalesce>` the `~DataQualityFlag.known` and
          `~DataQualityFlag.active` segment lists
        - forces the `active` segments to be a proper subset of the `known`
          segments

        .. note::

            this operations is performed in-place.

        Returns
        -------
        self
            a view of this flag, not a copy.
        """
        self.known = self.known.coalesce()
        self.active = self.active.coalesce()
        self.active = (self.known & self.active).coalesce()
        return self

    def __repr__(self):
        indent = " " * len("<%s(" % self.__class__.__name__)
        known = str(self.known).replace("\n",
                                        "\n%s      " % indent).split('\n')
        if len(known) > 10:
            known = known[:3] + ['%s      ...' % indent] + known[-3:]
        active = str(self.active).replace("\n",
                                          "\n%s       " % indent).split('\n')
        if len(active) > 10:
            active = active[:3] + ['%s        ...' % indent] + active[-3:]
        return ("<{1}({2},\n{0}known={3},\n{0}active={4},\n"
                "{0}description={5})>".format(indent, self.__class__.__name__,
                                              (self.name and repr(self.name) or
                                               'No name'),
                                              '\n'.join(known),
                                              '\n'.join(active),
                                              repr(self.description)))

    def copy(self):
        """Build an exact copy of this flag.

        Returns
        -------
        flag2 : `DataQualityFlag`
            a copy of the original flag, but with a fresh memory address.
        """
        new = self.__class__()
        new.ifo = self.ifo
        new.name = self.name
        new.version = self.version
        new.description = self.description
        new.known = self._ListClass([self._EntryClass(s[0], s[1]) for
                                    s in self.known])
        new.active = self._ListClass([self._EntryClass(s[0], s[1]) for
                                     s in self.active])
        return new

    def plot(self, **kwargs):
        """Plot this flag.

        Parameters
        ----------
        **kwargs
            all keyword arguments are passed to the
            :class:`~gwpy.plotter.segments.SegmentPlot` constructor.

        Returns
        -------
        plot : `~gwpy.plotter.segments.SegmentPlot`
            a new `Plot` with this flag displayed on a set of
            :class:`~gwpy.plotter.segments.SegmentAxes`.
        """
        from ..plotter import (rcParams, SegmentPlot)
        kwargs.setdefault('epoch', self.known[0][0])
        if self.label:
            kwargs.setdefault('label', self.label)
        elif rcParams['text.usetex']:
            kwargs.setdefault('label', self.texname)
        else:
            kwargs.setdefault('label', self.name)
        return SegmentPlot(self, **kwargs)

    def _parse_name(self, name):
        """Internal method to parse a `string` name into constituent
        `ifo, `name` and `version` components.

        Parameters
        ----------
        name : `str`
            the full name of a `DataQualityFlag` to parse, e.g.
            ``'H1:DMT-SCIENCE:1'``

        Returns
        -------
        (ifo, name, version)
            A tuple of component string parts

        Raises
        ------
        `ValueError`
            If the input ``name`` cannot be parsed into
            {ifo}:{tag}:{version} format.
        """
        if name is None:
            self.ifo = None
            self.tag = None
            self.version = None
        elif re_IFO_TAG_VERSION.match(name):
            match = re_IFO_TAG_VERSION.match(name).groupdict()
            self.ifo = match['ifo']
            self.tag = match['tag']
            self.version = int(match['version'])
        elif re_IFO_TAG.match(name):
            match = re_IFO_TAG.match(name).groupdict()
            self.ifo = match['ifo']
            self.tag = match['tag']
            self.version = None
        elif re_TAG_VERSION.match(name):
            match = re_TAG_VERSION.match(name).groupdict()
            self.ifo = None
            self.tag = match['tag']
            self.version = None
        else:
            raise ValueError("No flag name structure detected in '%s', flags "
                             "should be named as '{ifo}:{tag}:{version}'. "
                             "For arbitrary strings, use the "
                             "`DataQualityFlag.label` attribute" % name)
        return self.ifo, self.tag, self.version

    def __getslice__(self, slice_):
        return self.__class__(name=self.name, known=self.known,
                              active=self.active[slice_], label=self.label,
                              description=self.description, isgood=self.isgood)

    def __getitem__(self, item):
        if isinstance(item, int) or isinstance(item, slice):
            return self.active[item]
        else:
            self.__getslice__(item)

    def __and__(self, other):
        """Find the intersection of this one and ``other``.
        """
        return self.copy().__iand__(other)

    def __iand__(self, other):
        """Intersect this flag with ``other`` in-place.
        """
        self.known &= other.known
        self.active &= other.active
        return self

    def __sub__(self, other):
        """Find the difference between this flag and another.
        """
        return self.copy().__isub__(other)

    def __isub__(self, other):
        """Subtract the ``other`` `DataQualityFlag` from this one in-place.
        """
        self.active -= other.active
        return self

    def __or__(self, other):
        """Find the union of this flag and ``other``.
        """
        return self.copy().__ior__(other)

    def __ior__(self, other):
        """Add the ``other`` `DataQualityFlag` to this one in-place.
        """
        self.known |= other.known
        self.active |= other.active
        return self

    __add__ = __or__
    __iadd__ = __ior__


class _QueryDQSegDBThread(Thread):
    """Threaded DQSegDB query
    """
    def __init__(self, inqueue, outqueue, *args, **kwargs):
        Thread.__init__(self)
        self.in_ = inqueue
        self.out = outqueue
        self.args = args
        self.kwargs = kwargs

    def run(self):
        i, flag = self.in_.get()
        self.in_.task_done()
        try:
            self.out.put(
                (i, DataQualityFlag.query_dqsegdb(flag, *self.args,
                                                  **self.kwargs)))
        except Exception as e:
            self.out.put((i, e))
        self.out.task_done()


class DataQualityDict(OrderedDict):
    """An `~collections.OrderedDict` of (key, `DataQualityFlag`) pairs.

    Since the `DataQualityDict` is an `OrderedDict`, all iterations over
    its elements retain the order in which they were inserted.
    """
    _EntryClass = DataQualityFlag

    # -----------------------------------------------------------------------
    # classmethods

    @classmethod
    def query(cls, flag, *args, **kwargs):
        """Query for segments of a set of flags.

        This method intelligently selects the `~DataQualityDict.query_segdb`
        or the `~DataQualityDict.query_dqsegdb` methods based on the
        ``url`` kwarg given.

        Parameters
        ----------
        flags : `iterable`
            A list of flag names for which to query.
        *args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments
        url : `str`, optional, default: ``'https://segments.ligo.org'``
            URL of the segment database

        See Also
        --------
        DataQualityDict.query_segdb
        DataQualityDict.query_dqsegdb
            for details on the actual query engine, and documentation of
            other keyword arguments appropriate for each query

        Returns
        -------
        flag : `DataQualityFlag`
            A new `DataQualityFlag`, with the `known` and `active` lists
            filled appropriately.
        """
        url = kwargs.get('url', 'https://segments.ligo.org')
        if 'dqsegdb' in url or re.match('https://[a-z1-9-]+.ligo.org', url):
            return cls.query_dqsegdb(flag, *args, **kwargs)
        else:
            return cls.query_segdb(flag, *args, **kwargs)

    @classmethod
    def query_segdb(cls, flags, *args, **kwargs):
        """Query the inital LIGO segment database for a list of flags.

        Parameters
        ----------
        flags : `iterable`
            A list of flag names for which to query.
        *args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments.
        url : `str`, optional, default: ``'https://segments.ligo.org'``
            URL of the segment database.

        Returns
        -------
        flagdict : `DataQualityDict
            An ordered `DataQualityDict` of (name, `DataQualityFlag`)
            pairs.
        """
        # given segmentlist
        if len(args) == 1 and isinstance(args[0], SegmentList):
            qsegs = args[0]
        elif len(args) == 1 and len(args[0]) == 2:
            qsegs = SegmentList(Segment(map(to_gps, args[0])))
        elif len(args) == 2:
            qsegs = SegmentList([Segment(map(to_gps, args))])
        else:
            raise ValueError("DataQualityDict.query_segdb must be called with "
                             "a list of flag names, and either GPS start and "
                             "stop times, or a SegmentList of query segments")
        url = kwargs.pop('url', 'https://segments.ligo.org')
        if kwargs.pop('on_error', None) is not None:
            warnings.warn("DataQualityDict.query_segdb doesn't accept the "
                          "on_error keyword argument")
        if kwargs.keys():
            raise TypeError("DataQualityDict.query_segdb has no keyword "
                            "argument '%s'" % kwargs.keys()[0])
        # parse flags
        if isinstance(flags, (str, unicode)):
            flags = flags.split(',')
        else:
            flags = flags
        # process query
        from glue.segmentdb import (segmentdb_utils as segdb_utils,
                                    query_engine as segdb_engine)
        connection = segdb_utils.setup_database(url)
        engine = segdb_engine.LdbdQueryEngine(connection)
        segdefs = []
        for flag in flags:
            dqflag = DataQualityFlag(name=flag)
            ifo = dqflag.ifo
            name = dqflag.tag
            if dqflag.version is None:
                vers = '*'
            else:
                vers = dqflag.version
            for gpsstart, gpsend in qsegs:
                if gpsend == PosInfinity or float(gpsend) == +inf:
                    gpsend = to_gps('now').seconds
                gpsstart = float(gpsstart)
                if not gpsstart.is_integer():
                    raise ValueError("Segment database queries can only"
                                     "operate on integer GPS times")
                gpsend = float(gpsend)
                if not gpsend.is_integer():
                    raise ValueError("Segment database queries can only"
                                     "operate on integer GPS times")
                segdefs += segdb_utils.expand_version_number(
                    engine, (ifo, name, vers, gpsstart, gpsend, 0, 0))
        segs = segdb_utils.query_segments(engine, 'segment', segdefs)
        segsum = segdb_utils.query_segments(engine, 'segment_summary', segdefs)
        # build output
        out = cls()
        for definition, segments, summary in zip(segdefs, segs, segsum):
            # parse flag name
            flag = ':'.join(map(str, definition[:3]))
            name = flag.rsplit(':', 1)[0]
            # if versionless
            if flag.endswith('*'):
                flag = name
                key = name
            # if asked for versionless, but returned a version
            elif flag not in flags and name in flags:
                key = name
            # other
            else:
                key = flag
            # define flag
            if key not in out:
                out[key] = DataQualityFlag(name=flag)
            # add segments
            out[key].known.extend(summary)
            out[key].active.extend(segments)
        return out

    @classmethod
    def query_dqsegdb(cls, flags, *args, **kwargs):
        """Query the advanced LIGO DQSegDB for a list of flags.

        Parameters
        ----------
        flags : `iterable`
            A list of flag names for which to query.
        *args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments.
        on_error : `str`
            how to handle an error querying for one flag, one of

            - `'raise'` (default): raise the Exception
            - `'warn'`: print a warning
            - `'ignore'`: move onto the next flag as if nothing happened

        url : `str`, optional, default: ``'https://segments.ligo.org'``
            URL of the segment database.

        Returns
        -------
        flagdict : `DataQualityDict
            An ordered `DataQualityDict` of (name, `DataQualityFlag`)
            pairs.
        """
        # check on_error flag
        on_error = kwargs.pop('on_error', 'raise').lower()
        if on_error not in ['raise', 'warn', 'ignore']:
            raise ValueError("on_error must be one of 'raise', 'warn', "
                             "or 'ignore'")

        # set up threading
        inq = Queue()
        outq = Queue()
        for i in range(len(flags)):
            t = _QueryDQSegDBThread(inq, outq, *args, **kwargs)
            t.setDaemon(True)
            t.start()
        for i, flag in enumerate(flags):
            inq.put((i, flag))

        # capture output
        inq.join()
        outq.join()
        new = cls()
        results = zip(*sorted([outq.get() for i in range(len(flags))],
                              key=lambda x: x[0]))[1]
        for result, flag in zip(results, flags):
            if isinstance(result, Exception):
                new[flag] = cls._EntryClass(name=flag)
                result.args = ('%s [%s]' % (str(result), str(flag)),)
                if on_error == 'ignore':
                    pass
                elif on_error == 'warn':
                    warnings.warn(str(result))
                else:
                    raise result
            else:
                new[flag] = result
        return new

    # use input/output registry to allow multi-format reading
    read = classmethod(reader(doc="""
    Read segments from file into a `DataQualityDict`.

    Parameters
    ----------
    filename : `str`
        path of file to read
    format : `str`, optional
        source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.
    flags : `list`, optional, default: read all flags found
        list of flags to read, by default all flags are read separately.
    coalesce : `bool`, optional, default: `True`
        coalesce all `SegmentLists` before returning.

    Returns
    -------
    flagdict : `DataQualityDict`
        a new `DataQualityDict` of `DataQualityFlag` entries with ``active``
        and ``known`` segments seeded from the XML tables in the given
        file.

    Notes
    -----
    """))

    @classmethod
    def from_veto_definer_file(cls, fp, start=None, end=None, ifo=None,
                               format='ligolw'):
        """Read a `DataQualityDict` from a LIGO_LW XML VetoDefinerTable.
        """
        if start is not None:
            start = to_gps(start)
        if end is not None:
            end = to_gps(end)
        # read veto definer file
        from gwpy.table.lsctables import VetoDefTable
        veto_def_table = VetoDefTable.read(fp, format=format)
        # parse flag definitions
        out = cls()
        for row in veto_def_table:
            if ifo and row.ifo != ifo:
                continue
            if start and 0 < row.end_time < start:
                continue
            elif start:
                row.start_time = max(row.start_time, start)
            if end and row.start_time > end:
                continue
            elif end and not row.end_time:
                row.end_time = end
            elif end:
                row.end_time = min(row.end_time, end)
            flag = DataQualityFlag.from_veto_def(row)
            if flag.name in out:
                out[flag.name].known.extend(flag.known)
                out[flag.name].known.coalesce()
            else:
                out[flag.name] = flag
        return out

    # -----------------------------------------------------------------------
    # instance methods

    write = writer()

    def populate(self, source='https://segments.ligo.org',
                 segments=None, pad=True, **kwargs):
        """Query the segment database for each flag's active segments.

        This method assumes all of the metadata for each flag have been
        filled. Minimally, the following attributes must be filled

        .. autosummary::

           ~DataQualityFlag.name
           ~DataQualityFlag.known

        Segments will be fetched from the database, with any
        :attr:`~DataQualityFlag.padding` added on-the-fly.

        Entries in this dict will be modified in-place.

        Parameters
        ----------
        source : `str`
            source of segments for this flag. This must be
            either a URL for a segment database or a path to a file on disk.

        segments : `SegmentList`, optional
            a list of known segments during which to query, if not given,
            existing known segments for flags will be used.

        pad : `bool`, optional, default: `True`
            apply the `~DataQualityFlag.padding` associated with each
            flag, default: `True`.

        **kwargs
            any other keyword arguments to be passed to
            :meth:`DataQualityFlag.query` or :meth:`DataQualityFlag.read`.

        Returns
        -------
        self : `DataQualityDict`
            a reference to the modified DataQualityDict
        """
        # format source
        source = urlparse(source)
        # perform query for all segments
        segments = SegmentList(map(Segment, segments))
        if source.netloc:
            tmp = type(self).query(self.keys(), segments,
                                   url=source.geturl(), **kwargs)
        else:
            tmp = type(self).read(source.geturl(), self.name, **kwargs)
        # apply padding and wrap to given known segments
        for key in self:
            self[key].known &= tmp[key].known
            self[key].active = tmp[key].active
            if pad:
                self[key] = self[key].pad()
                self[key].known &= segments
                self[key].active &= segments
        return self

    def __iand__(self, other):
        for key, value in other.iteritems():
            if key in self:
                self[key] &= value
            else:
                self[key] = self._EntryClass()
        return self

    def __and__(self, other):
        if sum(len(s) for s in self.values()) <= sum(len(s) for s in
                                                     other.values()):
            return self.copy().__iand__(other)
        return other.copy().__iand__(self)

    def __ior__(self, other):
        for key, value in other.iteritems():
            if key in self:
                self[key] |= value
            else:
                self[key] = shallowcopy(value)
        return self

    def __or__(self, other):
        if (sum(len(s.active) for s in self.values()) >=
                sum(len(s.active) for s in other.values())):
            return self.copy().__ior__(other)
        return other.copy().__ior__(self)

    __iadd__ = __ior__
    __add__ = __or__

    def __isub__(self, other):
        for key, value in other.iteritems():
            if key in self:
                self[key] -= value
        return self

    def __sub__(self, other):
        return self.copy().__isub__(other)

    def __ixor__(self, other):
        for key, value in other.iteritems():
            if key in self:
                self[key] ^= value
            else:
                self[key] = shallowcopy(value)
        return self

    def __xor__(self, other):
        if sum(len(s) for s in self.values()) <= sum(len(s) for s in
                                                     other.values()):
            return self.copy().__ixor__(other)
        return other.copy().__ixor__(self)

    def __invert__(self):
        new = self.copy()
        for key, value in new.items():
            dict.__setitem__(new, key, ~value)
        return new

    def union(self):
        """Return the union of all flags in this dict

        Returns
        -------
        union : `DataQualityFlag`
            a new `DataQualityFlag` who's active and known segments
            are the union of those of the values of this dict
        """
        usegs = reduce(operator.or_, self.itervalues())
        usegs.name = ' | '.join(self.iterkeys())
        return usegs

    def intersection(self):
        """Return the intersection of all flags in this dict

        Returns
        -------
        intersection : `DataQualityFlag`
            a new `DataQualityFlag` who's active and known segments
            are the intersection of those of the values of this dict
        """
        isegs = reduce(operator.and_, self.itervalues())
        isegs.name = ' & '.join(self.iterkeys())
        return isegs

    def plot(self, label='key', **kwargs):
        """Plot the data for this dict.

        Parameters
        ----------
        label : `str`, optional
            labelling system to use, or fixed label for all flags,
            special values include

            - ``'key'``: use the key of the `DataQualityDict`,
            - ``'name'``: use the :attr:`~DataQualityFlag.name` of the flag

            If anything else, that fixed label will be used for all lines.

        **kwargs
            all other keyword arguments are passed to the plotter as
            appropriate

        See Also
        --------
        gwpy.plotter.SegmentPlot
        gwpy.plotter.SegmentAxes
        gwpy.plotter.SegmentAxes.plot_dqdict
        """
        from ..plotter import SegmentPlot
        figargs = dict()
        for key in ['figsize', 'dpi']:
            if key in kwargs:
                figargs[key] = kwargs.pop(key)
        axargs = dict()
        for key in ['insetlabels']:
            if key in kwargs:
                axargs[key] = kwargs.pop(key)
        plot_ = SegmentPlot(**figargs)
        ax = plot_.gca(**axargs)
        ax.plot(self, label=label, **kwargs)
        ax.autoscale(axis='y')
        return plot_



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2015)

# package metadata
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = '0.1b4'
__date__ = '2015-11-18 22:21:18 +0000'

# package version
version = '0.1b4'
major = 0
minor = 1
micro = None
debug = True
release = False

# repository version information
git_hash = '25914f2e2f31ba62ad3d2ea246697925ab4d5346'
git_tag = 'v0.1b4'
git_branch = 'master'
git_author = "Duncan Macleod <duncanmmacleod@gmail.com>"
git_committer = "Duncan Macleod <duncanmmacleod@gmail.com>"
git_is_dirty = False


#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting segments for a `StateVector`

I would like to examine the state of the internal seismic isolation system supporting the Fabry-Perot mirror at the end of the Y-arm at LHO, in order to investigate a noise source.

These data are private to the LIGO Scientific Collaboration and the Virgo Collaboration, but collaboration members can use the NDS2 service to download data.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.timeseries'

# First, we can import the `StateVector` class:
from gwpy.timeseries import StateVector

# Next, we define which bits we want to use, and can then
# :meth:`~StateVector.get` the data:
bits = [
    'Summary state',
    'State 1 damped',
    'Stage 1 isolated',
    'Stage 2 damped',
    'Stage 2 isolated',
    'Master switch ON',
    'Stage 1 WatchDog OK',
    'Stage 2 WatchDog OK',
]

data = StateVector.get('L1:ISI-ETMX_ODC_CHANNEL_OUT_DQ', 'May 22 2014 14:00', 'May 22 2014 15:00', bits=bits)
data = data.astype('uint32')  # hide

# For this example, we wish to :meth:`~StateVector.resample` the data to a
# much lower rate, to make visualising the state much easier:
data = data.resample(16)

# Finally, we make a :meth:`~StateVector.plot`, passing `insetlabels=True` to
# display the bit names inside the axes:
plot = data.plot(insetlabels=True)
plot.set_title('LLO ETMX internal seismic isolation state')
plot.add_bitmask('0b11101110')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""tconvert: a utility to convert to and from GPS times.

This method is inspired by the original tconvert utility, written by
Peter Shawhan.
"""

import datetime

from dateutil import parser as dateparser

from astropy.units import Quantity

from .. import version
from . import (Time, LIGOTimeGPS)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version
__all__ = ['tconvert', 'to_gps', 'from_gps']


def tconvert(gpsordate='now'):
    """Convert GPS times to ISO-format date-times and vice-versa.

    Parameters
    ----------
    gpsordate : `float`, `LIGOTimeGPS`, `Time`, `datetime.datetime`, ...
        input gps or date to convert

    Returns
    -------
    date : `datetime.datetime` or `LIGOTimeGPS`
        converted gps or date

    Notes
    -----
    If the input object is a `float` or `LIGOTimeGPS`,
    it will get converted from GPS format into a
    `datetime.datetime`, otherwise the input will be converted
    into `LIGOTimeGPS`.
    """
    # convert from GPS into datetime
    try:
        gps = LIGOTimeGPS(gpsordate)
    except (ValueError, TypeError):
        if hasattr(gpsordate, 'gpsSeconds'):
            return from_gps(gpsordate)
        else:
            return to_gps(gpsordate)
    else:
        return from_gps(gps)


def to_gps(t, *args, **kwargs):
    """Convert any input date/time into a `LIGOTimeGPS`.

    Any input object that can be cast as a
    `~astropy.time.Time` (with `str` going through the
    `datetime.datetime`) are acceptable.

    Parameters
    ----------
    t : `float`, `datetime.datetime`, `~astropy.time.Time`, `str`
        the input time, any object that can be converted into a
        `LIGOTimeGPS`, `~astropy.time.Time`, or `datetime.datetime`,
        is acceptable.
    *args, **kwargs
        other arguments to pass to pass to `~astropy.time.Time` if given

    Returns
    -------
    gps : `LIGOTimeGPS`
        the number of GPS seconds (non-integer) since the start of the
        epoch (January 6 1980).

    Raises
    ------
    TypeError
        if a `str` input cannot be parsed as a `datetime.datetime`.
    ValueError
        if the input cannot be cast as a `~astropy.time.Time` or
        `LIGOTimeGPS`.
    """
    # allow Time conversion to override type-checking
    if args or kwargs:
        return Time(t, *args, **kwargs).utc.gps
    # if lal.LIGOTimeGPS
    if hasattr(t, 'gpsSeconds'):
        return LIGOTimeGPS(t.gpsSeconds, t.gpsNanoSeconds)
    # or convert numeric string to float (e.g. '123.456')
    try:
        t = float(t)
    except (TypeError, ValueError):
        pass
    # or convert str into datetime.datetime
    if isinstance(t, str):
        t = str_to_datetime(t)
    # or convert tuple into datetime.datetime
    elif isinstance(t, (tuple, list)):
        t = datetime.datetime(*t)
    # and then into lal.LIGOTimeGPS or Time
    if isinstance(t, datetime.date):
        # try and use LAL, it's more reliable (possibly)
        try:
            from lal import UTCToGPS
        except ImportError:
            if not isinstance(t, datetime.datetime):
                t = datetime.datetime.combine(t, datetime.time.min)
            t = Time(t, scale='utc')
        else:
            return to_gps(UTCToGPS(t.timetuple()))
    # and then into LIGOTimeGPS
    if isinstance(t, Time):
        return time_to_gps(t)
    # extract quantity to a float in seconds
    if isinstance(t, Quantity):
        t = t.to('second').value
    # if all else fails...
    return LIGOTimeGPS(t)


def from_gps(gps):
    """Convert a GPS time into a `datetime.datetime`.

    Parameters
    ----------
    gps : `LIGOTimeGPS`
        GPS time to convert

    Returns
    -------
    datetime : `datetime.datetime`
        ISO-format datetime equivalent of input GPS time
    """
    try:
        gps = LIGOTimeGPS(gps)
    except (ValueError, TypeError):
        gps = LIGOTimeGPS(float(gps))
    dt = Time(gps.seconds, gps.nanoseconds * 1e-9,
              format='gps', scale='utc').datetime
    if float(gps).is_integer():
        return dt.replace(microsecond=0)
    else:
        return dt


def time_to_gps(t):
    """Convert a `Time` into `LIGOTimeGPS`.

    Parameters
    ----------
    t : `~astropy.time.Time`
        formatted `Time` object to convert

    Returns
    -------
    gps : `LIGOTimeGPS`
        Nano-second precision `LIGOTimeGPS` time
    """
    t = t.utc
    dt = t.datetime
    gps = t.gps
    if ((isinstance(dt, datetime.datetime) and not dt.microsecond) or
            type(dt) is datetime.date):
        gps = int(gps)
    return LIGOTimeGPS(gps)


def str_to_datetime(datestr):
    """Convert a `str` representing a datetime into a `datetime.datetime`.

    Parameters
    ----------
    datestr : `str`
        date-like string parseable by :meth:`dateutil.parser.parse`, or
        one of the following special cases

        - 'now' : second precision for current time
        - 'today'
        - 'tomorrow'
        - 'yesterday'

    Returns
    -------
    date : `datetime.datetime`
        `datetime.datetime` version of the input ``datestr``

    Raises
    ------
    TypeError
        if ``datestr`` cannot be parsed by :meth:`dateutil.parser.parse`
    """
    datestr = str(datestr).lower()
    if datestr == 'now':
        date = datetime.datetime.utcnow().replace(microsecond=0)
    elif datestr == 'today':
        date = datetime.date.today()
    elif datestr == 'tomorrow':
        today = datetime.date.today()
        date = today + datetime.timedelta(days=1)
    elif datestr == "yesterday":
        today = datetime.date.today()
        date = today - datetime.timedelta(days=1)
    else:
        try:
            date = dateparser.parse(datestr)
        except TypeError as e:
            e.args = ("Cannot parse date string %r: %s"
                      % (datestr, e.args[0]),)
            raise
    return date



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Docstring
"""

from __future__ import division

from math import log10

import numpy

from matplotlib.cbook import iterable
from matplotlib.projections import register_projection

from glue import iterutils
from glue.ligolw.table import Table

from ..table.utils import get_table_column
from .axes import Axes
from .core import Plot
from ..data import Series
from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class HistogramAxes(Axes):
    """A set of `Axes` on which to display a histogram.
    """
    name = 'histogram'

    def hist_series(self, series, **kwargs):
        """Add a histogram of the given 1-dimensional series to these `Axes`.

        Parameters
        ----------
        series : :class:`~gwpy.data.series.Series`
            1-dimensional `Series` of data.
        **kwargs
            common histogram keyword arguments to pass to
            :meth:`~HistogramAxes.hist`.

        See Also
        --------
        HistogramAxes.hist : for details on keyword arguments
        """
        return self.hist(series.value, **kwargs)

    def hist_table(self, table, column, **kwargs):
        """Add a histogram of the given :class:`~glue.ligolw.table.Table`.

        Parameters
        ----------
        table : :class:`~glue.ligolw.table.Table`
            LIGO_LW-format table of data to analyse.
        column : `str`
            name of ``table`` column to histogram.
        **kwargs
            common histogram keyword arguments to pass to
            :meth:`~HistogramAxes.hist`.

        See Also
        --------
        HistogramAxes.hist : for details on keyword arguments
        """
        data = get_table_column(table, column)
        return self.hist(data, **kwargs)

    def hist(self, x, **kwargs):
        if iterable(x) and len(x) == 0:
            x = numpy.ndarray((0,))
        logbins = kwargs.pop('logbins', False)
        bins = kwargs.get('bins', 30)
        weights = kwargs.get('weights', None)
        if isinstance(weights, (float, int)):
            kwargs['weights'] = []
            if isinstance(x, numpy.ndarray) or not iterable(x[0]):
                kwargs['weights'].append(numpy.ones_like(x) * weights)
            else:
                for x2 in x:
                    kwargs['weights'].append(numpy.ones_like(x2) * weights)
        if logbins and (bins is None or isinstance(bins, (float, int))):
            bins = bins or 30
            range_ = kwargs.pop('range', self.common_limits(x))
            if range_[1] == range_[0]:
                kwargs['bins'] = bins
            else:
                kwargs['bins'] = self.bin_boundaries(range_[0], range_[1],
                                                     bins, log=True)
        if kwargs.get('histtype', None) == 'stepfilled':
            kwargs.setdefault('edgecolor', 'black')
        return super(HistogramAxes, self).hist(x, **kwargs)
    hist.__doc__ = Axes.hist.__doc__

    @staticmethod
    def common_limits(datasets, default_min=0, default_max=0):
        """Find the global maxima and minima of a list of datasets.

        Parameters
        ----------
        datasets : `iterable`
            list (or any other iterable) of data arrays to analyse.
        default_min : `float`, optional
            fall-back minimum value if datasets are all empty.
        default_max : `float`, optional
            fall-back maximum value if datasets are all empty.

        Returns
        -------
        (min, max) : `float`
            2-tuple of common minimum and maximum over all datasets.
        """
        if isinstance(datasets, numpy.ndarray) or not iterable(datasets[0]):
            datasets = [datasets]
        max_stat = max(list(iterutils.flatten(datasets)) + [-numpy.inf])
        min_stat = min(list(iterutils.flatten(datasets)) + [numpy.inf])
        if numpy.isinf(-max_stat):
            max_stat = default_max
        if numpy.isinf(min_stat):
            min_stat = default_min
        return min_stat, max_stat

    @staticmethod
    def bin_boundaries(lower, upper, num, log=False):
        """Determine the bin boundaries for the given interval.

        The returned array contains the left edge of all bins, as well as
        the right-most edge of the right-most bin.

        Parameters
        ----------
        lower : `float`
            minimum boundary for bins. This value will be the first
            element of the returned bins array.
        upper : `float`
            maximum boundary for bins. This value will be the last
            element of the returned bins array.
        num : `int`
            number of bins to generate
        log : `bool`, optional, default: `False`
            if `True` generate logarithmically-spaced bins, else
            generate linearly-spaced bins.

        Returns
        -------
        bins : :class:`~numpy.ndarray`
            array of bins (length ``len(num)`` + 1).
        """
        if log:
            return numpy.logspace(log10(lower), log10(upper), num+1,
                                  endpoint=True)
        else:
            return numpy.linspace(lower, upper, num+1, endpoint=True)

register_projection(HistogramAxes)


class HistogramPlot(Plot):
    """A plot showing a histogram of data
    """
    _DefaultAxesClass = HistogramAxes

    def __init__(self, *data, **kwargs):
        """Generate a new `HistogramPlot` from some ``data``.
        """
        # extract histogram arguments
        histargs = dict()
        for key in ['bins', 'range', 'normed', 'weights', 'cumulative',
                    'bottom', 'histtype', 'align', 'orientation', 'rwidth',
                    'log', 'color', 'label', 'stacked', 'logbins']:
            try:
                histargs[key] = kwargs.pop(key)
            except KeyError:
                pass
        # generate Figure
        super(HistogramPlot, self).__init__(**kwargs)
        # plot data
        if data:
            ax = self.gca()
            data = list(data)
        else:
            ax = None
        while data:
            dataset = data.pop(0)
            if isinstance(dataset, Series):
                ax.hist_series(dataset, **histargs)
            elif isinstance(dataset, Table):
                column = data.pop()
                ax.hist_table(dataset, column, **histargs)
            else:
                ax.hist(dataset, **histargs)
        if ax and histargs.get('logbins', False):
            if histargs.get('orientation', 'vertical') == 'vertical':
                ax.set_xscale('log')
            else:
                ax.set_yscale('log')



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting events as 2-d tiles

I would like to study the event triggers generated by the `ExcessPower <link>`_
gravitational-wave burst detection algorithm, over a small stretch of data.

The data from which these events were generated are a simulation of Gaussian noise
with the Advanced LIGO design spectrum, and so don't actually contain any real
gravitational waves, but will help tune the algorithm to improve detection of
future, real signals.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.table.lsctables'

# First, we import the `SnglBurstTable` class
from gwpy.table.lsctables import SnglBurstTable

# and read the table of events:
events = SnglBurstTable.read('../../gwpy/tests/data/H1-LDAS_STRAIN-968654552-10.xml.gz')

# We can make a plot of these events as 2-dimensional tiles by specifying
# the x- and y-axis columns, and the widths in those directions:
plot = events.plot('time', 'central_freq', 'duration', 'bandwidth', color='snr', epoch=968654552)
plot.set_xlim(968654552, 968654552+10)
plot.set_ylabel('Frequency [Hz]')
plot.set_yscale('log')
plot.set_title('LIGO Hanford Observatory event triggers for GW100916')
plot.add_colorbar(clim=[1, 5], label='Signal-to-noise ratio', cmap='hot_r')
plot.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

""" Spectrum plots
"""
from cliproduct import CliProduct


class Spectrum(CliProduct):

    def get_action(self):
        """Return the string used as "action" on command line."""
        return 'spectrum'

    def init_cli(self, parser):
        """Set up the argument list for this product"""
        self.arg_chan1(parser)
        self.arg_freq(parser)
        self.arg_ax_xlf(parser)
        self.arg_ax_logy(parser)
        self.arg_plot(parser)
        self.xaxis_is_freq = True
        return

    def get_ylabel(self, args):
        """Text for y-axis label"""
        if args.nology:
            ylabel = r'$\mathrm{log_{10}  ASD}$ ' \
                     r'$\left( \frac{\mathrm{Counts}}' \
                     r'{\sqrt{\mathrm{Hz}}}\right)$'
        else:
            ylabel = r'$\mathrm{ASD}$ $\left( \frac{\mathrm{Counts}}' \
                     r'{\sqrt{\mathrm{Hz}}}\right)$'
        return ylabel

    def get_title(self):
        """Start of default super title, first channel is appended to it"""
        return 'Spectrum: '

    def get_xlabel(self):
        xlabel = 'Frequency (Hz)'
        return xlabel

    def freq_is_y(self):
        """This plot puts frequency on the y-axis of the graph"""
        return False

    def gen_plot(self, arg_list):
        """Generate the plot from time series and arguments"""
        self.is_freq_plot = True

        fftlen = 1.0
        if arg_list.secpfft:
            fftlen = float(arg_list.secpfft)
        self.secpfft = fftlen
        ovlap = 0.5
        if arg_list.overlap:
            ovlap = float(arg_list.overlap)
        self.overlap = ovlap

        self.log(2, "Calculating spectrum secpfft: %.2f, overlap: %.2f" %
                 (fftlen, ovlap))
        spectra = []

        # calculate and plot the first spectrum
        spectrum = self.timeseries[0].asd(fftlen, fftlen*ovlap)
        spectra.append(spectrum)

        fs = self.timeseries[0].sample_rate.value
        self.fmin = 1/self.secpfft
        self.fmax = fs/2
        self.ymin = spectrum.value.min()
        self.ymax = spectrum.value.max()

        label = self.timeseries[0].channel.name
        if len(self.start_list) > 1:
            label += ", %s" % self.timeseries[0].epoch.gps
        spectrum.name = label
        self.plot = spectrum.plot()

        # if we have more time series calculate and add to the first plot
        if len(self.timeseries) > 1:
            for idx in range(1, len(self.timeseries)):
                specb = self.timeseries[idx].asd(fftlen, ovlap*fftlen)
                spectra.append(specb)
                fsb = self.timeseries[idx].sample_rate.value
                self.fmax = max(self.fmax, fsb/2)
                self.ymin = min(self.ymin, specb.value.min())
                self.ymax = max(self.ymax, specb.value.max())

                label = self.timeseries[idx].channel.name
                if len(self.start_list) > 1:
                    label += ", %s" % self.timeseries[idx].epoch.gps
                specb.name = label
                self.plot.add_spectrum(specb)
        self.log(2, ('Frequency range: [%f, %f]' % (self.fmin, self.fmax)))
        # if the specified frequency limits adjust our ymin and ymax values
        # at this point self.ymin and self.ymax represent the full spectra
        if arg_list.fmin or arg_list.fmax:
            import numpy
            mymin = self.ymax   # guaranteed to be >= anything we look at
            mymax = self.ymin   # guaranteed to be <= anything we look at
            myfmin = self.fmin
            myfmax = self.fmax
            if arg_list.fmin:
                myfmin = float(arg_list.fmin)
            if arg_list.fmax:
                myfmax = float(arg_list.fmax)

            for idx in range(0, len(spectra)):
                t = numpy.where(spectra[idx].frequencies.value >= myfmin)
                if t[0].size:
                    strt = t[0][0]
                    t = numpy.where(spectra[idx].frequencies.value >= myfmax)
                    if t[0].size:
                        stop = t[0][0]
                    else:
                        stop = spectra[idx].frequencies.size - 1
                    mymin = min(mymin, numpy.min(spectra[idx].value[strt:stop]))
                    mymax = max(mymax, numpy.max(spectra[idx].value[strt:stop]))

            self.ymin = mymin
            self.ymax = mymax
        return



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Methods to calculate a rate TimeSeries from a LIGO_LW Table.
"""

import operator as _operator
from math import ceil

import numpy

from .. import version
from .utils import (EVENT_TABLES, get_table_column)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version
__all__ = ['event_rate', 'binned_event_rates']

OPERATORS = {'<': _operator.lt, '<=': _operator.le, '=': _operator.eq,
             '>=': _operator.ge, '>': _operator.gt, '==': _operator.is_,
             '!=': _operator.is_not}


def event_rate(self, stride, start=None, end=None, timecolumn='time'):
    """Calculate the rate `~gwpy.timeseries.TimeSeries` for this `Table`.

    Parameters
    ----------
    stride : `float`
        size (seconds) of each time bin
    start : `float`, :class:`~gwpy.time.LIGOTimeGPS`, optional
        GPS start epoch of rate :class:`~gwpy.timeseries.TimeSeries`
    end : `float`, :class:`~gwpy.time.LIGOTimeGPS`, optional
        GPS end time of rate :class:`~gwpy.timeseries.TimeSeries`.
        This value will be rounded up to the nearest sample if needed.
    timecolumn : `str`, optional, default: ``time``
        name of time-column to use when binning events

    Returns
    -------
    rate : :class:`~gwpy.timeseries.TimeSeries`
        a `TimeSeries` of events per second (Hz)
    """
    from gwpy.timeseries import TimeSeries
    # get time data
    times = get_table_column(self, timecolumn)
    # generate time bins
    if not start:
        start = times.min()
    if not end:
        end = times.max()
    nsamp = int(ceil((end - start) / stride))
    timebins = numpy.arange(nsamp + 1) * stride + start
    # histogram data and return
    out = TimeSeries(numpy.histogram(times, bins=timebins)[0] / float(stride),
                     epoch=start, sample_rate=1/float(stride), unit='Hz',
                     name='Event rate')
    return out


def binned_event_rates(self, stride, column, bins, operator='>=',
                       start=None, end=None, timecolumn='time'):
    """Calculate an event rate `~gwpy.timeseries.TimeSeriesDict` over
    a number of bins.

    Parameters
    ----------
    stride : `float`
        size (seconds) of each time bin
    column : `str`
        name of column by which to bin.
    bins : `list`
        a list of `tuples <tuple>` marking containing bins, or a list of
        `floats <float>` defining bin edges against which an math operation
        is performed for each event.
    operator : `str`, `callable`
        one of:

        - ``'<'``, ``'<='``, ``'>'``, ``'>='``, ``'=='``, ``'!='``,
          for a standard mathematical operation,
        - ``'in'`` to use the list of bins as containing bin edges, or
        - a callable function that takes compares an event value
          against the bin value and returns a boolean.

        .. note::

           If ``bins`` is given as a list of tuples, this argument
           is ignored.

    start : `float`, :class:`~gwpy.time.LIGOTimeGPS`, optional
        GPS start epoch of rate `~gwpy.timeseries.TimeSeries`.
    end : `float`, `~gwpy.time.LIGOTimeGPS`, optional
        GPS end time of rate `~gwpy.timeseries.TimeSeries`.
        This value will be rounded up to the nearest sample if needed.
    timecolumn : `str`, optional, default: ``time``
        name of time-column to use when binning events

    Returns
    -------
    rates : :class:`~gwpy.timeseries.TimeSeriesDict`
        a dict of (bin, `~gwpy.timeseries.TimeSeries`) pairs describing a
        rate of events per second (Hz) for each of the bins.
    """
    from gwpy.timeseries import (TimeSeries, TimeSeriesDict)
    from gwpy.plotter.table import get_column_string
    # get time data
    times = get_table_column(self, timecolumn)

    # get channel
    try:
        channel = self[0].channel
    except (IndexError, AttributeError):
        channel = None

    # generate time bins
    if not start:
        start = times.min()
    if not end:
        end = times.max()
    nsamp = int(ceil((end - start) / stride))
    timebins = numpy.arange(nsamp + 1) * stride + start
    # generate column bins
    if not bins:
        bins = [(-numpy.inf, numpy.inf)]
    if operator == 'in' and not isinstance(bins[0], tuple):
        bins2 = []
        for i, bin_ in enumerate(bins[:-1]):
            bins2.append((bin_, bins[i+1]))
        bins = bins2
    elif isinstance(operator, (unicode, str)):
        op = OPERATORS[operator]
    else:
        op = operator
    coldata = get_table_column(self, column)
    colstr = get_column_string(column)
    # generate one TimeSeries per bin
    out = TimeSeriesDict()
    for bin_ in bins:
        if isinstance(bin_, tuple):
            bintimes = times[(coldata >= bin_[0]) & (coldata < bin_[1])]
        else:
            bintimes = times[op(coldata, bin_)]
        out[bin_] = TimeSeries(
            numpy.histogram(bintimes, bins=timebins)[0] / float(stride),
            epoch=start, sample_rate=1/float(stride), unit='Hz',
            name='%s $%s$ %s' % (colstr, operator, bin_), channel=channel)
    return out


# attach methods to lsctables
for table in EVENT_TABLES:
    table.event_rate = event_rate
    table.binned_event_rates = binned_event_rates



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Spectrogram object
"""

import warnings

import numpy

import scipy
from scipy import signal
from astropy import units

from ..detector import Channel
from ..data import (Array2D, Series)
from ..segments import Segment
from ..timeseries import (TimeSeries, TimeSeriesList)
from ..spectrum import Spectrum
from ..utils.docstring import interpolate_docstring

from .. import version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org"
__version__ = version.version

__all__ = ['Spectrogram', 'SpectrogramList']


@interpolate_docstring
class Spectrogram(Array2D):
    """A 2D array holding a spectrogram of time-frequency data

    Parameters
    ----------
    %(Array1)s

    %(time-axis)s

    %(frequency-axis)s

    %(Array2)s

    Notes
    -----
    Key methods:

    .. autosummary::

       ~Spectrogram.read
       ~Spectrogram.write
       ~Spectrogram.plot
       ~Spectrogram.zpk
    """
    _metadata_slots = ['name', 'channel', 'epoch', 'dt', 'f0', 'df']
    _default_xunit = TimeSeries._default_xunit
    _default_yunit = Spectrum._default_xunit

    def __new__(cls, data, unit=None, name=None, channel=None, epoch=None,
                dt=None, times=None, f0=None, df=None, frequencies=None,
                **kwargs):
        """Generate a new Spectrogram.
        """
        # parse Channel input
        if isinstance(channel, Channel):
            name = name or channel.name
            unit = unit or channel.unit
        # get axis-based params
        if epoch is None:
            epoch = kwargs.pop('x0', 0)
        if dt is None:
            dt = kwargs.pop('dx', 1)
        if f0 is None:
            f0 = kwargs.pop('y0', 0)
        if df is None:
            df = kwargs.pop('dy', 1)
        if times is None:
            times = kwargs.pop('xindex', None)
        if frequencies is None:
            frequencies = kwargs.pop('frequencies', None)
        # generate Spectrogram
        new = super(Spectrogram, cls).__new__(cls, data, unit=unit, name=name,
                                              channel=channel, y0=f0, dx=dt,
                                              dy=df, xindex=times,
                                              yindex=frequencies, **kwargs)
        if epoch is not None:
            new.epoch = epoch
        return new

    # -------------------------------------------
    # Spectrogram properties

    epoch = property(TimeSeries.epoch.__get__, TimeSeries.epoch.__set__,
                     TimeSeries.epoch.__delete__,
                     """Starting GPS epoch for this `Spectrogram`

                     :type:`~gwpy.segments.Segment`
                     """)

    span = property(TimeSeries.span.__get__, TimeSeries.span.__set__,
                    TimeSeries.span.__delete__,
                    """GPS [start, stop) span for this `Spectrogram`

                    :type:`~gwpy.segments.Segment`
                    """)

    dt = property(TimeSeries.dt.__get__, TimeSeries.dt.__set__,
                  TimeSeries.dt.__delete__,
                  """Time-spacing for this `Spectrogram`

                  :type:`~astropy.units.Quantity` in seconds
                  """)

    f0 = property(Array2D.y0.__get__, Array2D.y0.__set__,
                  Array2D.y0.__delete__,
                  """Starting frequency for this `Spectrogram`

                  :type: `~astropy.units.Quantity` in Hertz
                  """)

    df = property(Array2D.dy.__get__, Array2D.dy.__set__,
                  Array2D.dy.__delete__,
                  """Frequency spacing of this `Spectrogram`

                  :type: `~astropy.units.Quantity` in Hertz
                  """)

    times = property(fget=Array2D.xindex.__get__,
                     fset=Array2D.xindex.__set__,
                     fdel=Array2D.xindex.__delete__,
                     doc="""Series of GPS times for each sample""")

    frequencies = property(fget=Array2D.yindex.__get__,
                           fset=Array2D.yindex.__set__,
                           fdel=Array2D.yindex.__delete__,
                           doc="Series of frequencies for this Spectrogram")

    band = property(fget=Array2D.yspan.__get__,
                    fset=Array2D.yspan.__set__,
                    fdel=Array2D.yspan.__delete__,
                    doc="""Frequency band described by this `Spectrogram`""")

    # -------------------------------------------
    # Spectrogram methods

    def ratio(self, operand):
        """Calculate the ratio of this `Spectrogram` against a reference

        Parameters
        ----------
        operand : `str`, `~gwpy.spectrum.Spectrum`, `~astropy.units.Quantity`
            `Spectrum` or `Quantity` against which to weight, or one of

            - ``'mean'`` : weight against the mean of each spectrum
              in this Spectrogram
            - ``'median'`` : weight against the median of each spectrum
              in this Spectrogram

        Returns
        -------
        spectrogram : `Spectrogram`
            a new `Spectrogram`
        """
        if operand == 'mean':
            operand = self.mean(axis=0)
        elif operand == 'median':
            operand = self.median(axis=0)
        elif isinstance(operand, str):
            raise ValueError("operand %r unrecognised, please give a Quantity"
                             " or one of: 'mean', 'median'" % operand)
        out = self / operand
        return out

    def plot(self, **kwargs):
        """Plot the data for this `Spectrogram`
        """
        from ..plotter import SpectrogramPlot
        return SpectrogramPlot(self, **kwargs)

    @classmethod
    def from_spectra(cls, *spectra, **kwargs):
        """Build a new `Spectrogram` from a list of spectra.

        Parameters
        ----------
        *spectra
            any number of `~gwpy.spectrum.Spectrum` series
        dt : `float`, `~astropy.units.Quantity`, optional
            stride between given spectra

        Returns
        -------
        Spectrogram
            a new `Spectrogram` from a vertical stacking of the spectra
            The new object takes the metadata from the first given
            `~gwpy.spectrum.Spectrum` if not given explicitly

        Notes
        -----
        Each `~gwpy.spectrum.Spectrum` passed to this
        constructor must be the same length.
        """
        data = numpy.vstack([s.value for s in spectra])
        s1 = spectra[0]
        if not all(s.f0 == s1.f0 for s in spectra):
            raise ValueError("Cannot stack spectra with different f0")
        if not all(s.df == s1.df for s in spectra):
            raise ValueError("Cannot stack spectra with different df")
        kwargs.setdefault('name', s1.name)
        kwargs.setdefault('epoch', s1.epoch)
        kwargs.setdefault('f0', s1.f0)
        kwargs.setdefault('df', s1.df)
        kwargs.setdefault('unit', s1.unit)
        if not ('dt' in kwargs or 'times' in kwargs):
            try:
                kwargs.setdefault('dt', spectra[1].epoch.gps - s1.epoch.gps)
            except AttributeError:
                raise ValueError("Cannot determine dt (time-spacing) for "
                                 "Spectrogram from inputs")
        return Spectrogram(data, **kwargs)

    def percentile(self, percentile):
        """Calculate a given spectral percentile for this `Spectrogram`.

        Parameters
        ----------
        percentile : `float`
            percentile (0 - 100) of the bins to compute

        Returns
        -------
        spectrum : `~gwpy.spectrum.Spectrum`
            the given percentile `Spectrum` calculated from this
            `SpectralVaraicence`
        """
        out = scipy.percentile(self.value, percentile, axis=0)
        name = '%s %s%% percentile' % (self.name, percentile)
        return Spectrum(out, epoch=self.epoch, channel=self.channel,
                        name=name, f0=self.f0, df=self.df,
                        frequencies=(hasattr(self, '_frequencies') and
                                     self.frequencies or None))

    def zpk(self, zeros, poles, gain):
        """Filter this `Spectrogram` by applying a zero-pole-gain filter

        Parameters
        ----------
        zeros : `array-like`
            list of zero frequencies (in Hertz)
        poles : `array-like`
            list of pole frequencies (in Hertz)
        gain : `float`
            DC gain of filter

        Returns
        -------
        specgram : `Spectrogram`
            the frequency-domain filtered version of the input data

        See Also
        --------
        Spectrogram.filter
            for details on how a digital ZPK-format filter is applied

        Examples
        --------
        To apply a zpk filter with file poles at 100 Hz, and five zeros at
        1 Hz (giving an overall DC gain of 1e-10)::

            >>> data2 = data.zpk([100]*5, [1]*5, 1e-10)
        """
        return self.filter(zeros, poles, gain)

    def filter(self, *filt, **kwargs):
        """Apply the given filter to this `Spectrogram`.

        Recognised filter arguments are converted into the standard
        ``(numerator, denominator)`` representation before being applied
        to this `Spectrogram`.

        .. note::

           Unlike the related
           :meth:`TimeSeries.filter <gwpy.timeseries.TimeSeries.filter>`
           method, here all frequency information (e.g. frequencies of
           poles or zeros in a ZPK) is assumed to be in Hertz.

        Parameters
        ----------
        *filt
            one of:

            - `scipy.signal.lti`
            - ``(numerator, denominator)`` polynomials
            - ``(zeros, poles, gain)``
            - ``(A, B, C, D)`` 'state-space' representation

        Returns
        -------
        result : `Spectrogram`
            the filtered version of the input `Spectrogram`

        See also
        --------
        Spectrum.zpk
            for information on filtering in zero-pole-gain format
        scipy.signal.zpk2tf
            for details on converting ``(zeros, poles, gain)`` into
            transfer function format
        scipy.signal.ss2tf
            for details on converting ``(A, B, C, D)`` to transfer function
            format
        scipy.signal.freqs
            for details on the filtering calculation

        Raises
        ------
        ValueError
            If ``filt`` arguments cannot be interpreted properly
        """
        # parse filter
        if len(filt) == 1 and isinstance(filt[0], signal.lti):
            filt = filt[0]
            a = filt.den
            b = filt.num
        elif len(filt) == 2:
            b, a = filt
        elif len(filt) == 3:
            b, a = signal.zpk2tf(*filt)
        elif len(filt) == 4:
            b, a = signal.ss2tf(*filt)
        else:
            raise ValueError("Cannot interpret filter arguments. Please give "
                             "either a signal.lti object, or a tuple in zpk "
                             "or ba format. See scipy.signal docs for "
                             "details.")
        if isinstance(a, float):
            a = numpy.array([a])
        # parse keyword args
        inplace = kwargs.pop('inplace', False)
        if kwargs:
            raise TypeError("Spectrogram.filter() got an unexpected keyword "
                            "argument '%s'" % list(kwargs.keys())[0])
        f = self.frequencies.value.copy()
        if f[0] == 0:
            f[0] = 1e-100
        fresp = numpy.nan_to_num(abs(signal.freqs(b, a, f)[1]))
        if inplace:
            self *= fresp
            return self
        else:
            new = self * fresp
            return new

    def variance(self, bins=None, low=None, high=None, nbins=500,
                 log=False, norm=False, density=False):
        """Calculate the `SpectralVariance` of this `Spectrogram`.

        Parameters
        ----------
        bins : `~numpy.ndarray`, optional, default `None`
            array of histogram bin edges, including the rightmost edge
        low : `float`, optional, default: `None`
            left edge of lowest amplitude bin, only read
            if ``bins`` is not given
        high : `float`, optional, default: `None`
            right edge of highest amplitude bin, only read
            if ``bins`` is not given
        nbins : `int`, optional, default: `500`
            number of bins to generate, only read if ``bins`` is not
            given
        log : `bool`, optional, default: `False`
            calculate amplitude bins over a logarithmic scale, only
            read if ``bins`` is not given
        norm : `bool`, optional, default: `False`
            normalise bin counts to a unit sum
        density : `bool`, optional, default: `False`
            normalise bin counts to a unit integral

        Returns
        -------
        specvar : `SpectralVariance`
            2D-array of spectral frequency-amplitude counts

        See Also
        --------
        :func:`numpy.histogram`
            for details on specifying bins and weights
        """
        from ..spectrum import SpectralVariance
        return SpectralVariance.from_spectrogram(
            self, bins=bins, low=low, high=high, nbins=nbins, log=log,
            norm=norm, density=density)

    # -------------------------------------------
    # connectors

    def is_compatible(self, other):
        """Check whether metadata attributes for self and other match.
        """
        if type(other) in [list, tuple, numpy.ndarray]:
            return True
        if not numpy.isclose(
                self.dt.decompose().value, other.dt.decompose().value):
            raise ValueError("Spectrogram time resolutions do not match: "
                             "%s vs %s." % (self.dt, other.dt))
        if not numpy.isclose(
                self.df.decompose().value, other.df.decompose().value):
            raise ValueError("Spectrogram frequency resolutions do not match:"
                             "%s vs %s." % (self.df, other.df))
        if not numpy.isclose(
                self.f0.decompose().value, other.f0.decompose().value):
            raise ValueError("Spectrogram starting frequencies do not match:"
                             "%s vs %s." % (self.f0, other.f0))
        if not self.unit == other.unit:
            raise ValueError("Spectrogram units do not match: %s vs %s."
                             % (self.unit, other.unit))
        return True

    def crop_frequencies(self, low=None, high=None, copy=False):
        """Crop this `Spectrogram` to the specified frequencies

        Parameters
        ----------
        low : `float`
            lower frequency bound for cropped `Spectrogram`
        high : `float`
            upper frequency bound for cropped `Spectrogram`
        copy : `bool`
            if `False` return a view of the original data, otherwise create
            a fresh memory copy

        Returns
        -------
        spec : `Spectrogram`
            A new `Spectrogram` with a subset of data from the frequency
            axis
        """
        if low is not None:
            low = units.Quantity(low, self._default_yunit)
        if high is not None:
            high = units.Quantity(high, self._default_yunit)
        # check low frequency
        if low is not None and low == self.f0:
            low = None
        elif low is not None and low < self.f0:
            warnings.warn('Spectrogram.crop_frequencies given low frequency '
                          'cutoff below f0 of the input Spectrogram. Low '
                          'frequency crop will have no effect.')
        # check high frequency
        if high is not None and high.value == self.band[1]:
            high = None
        elif high is not None and high.value > self.band[1]:
            warnings.warn('Spectrogram.crop_frequencies given high frequency '
                          'cutoff above cutoff of the input Spectrogram. High '
                          'frequency crop will have no effect.')
        # find low index
        if low is None:
            idx0 = None
        else:
            idx0 = int((low.value - self.f0.value) // self.df.value)
        # find high index
        if high is None:
            idx1 = None
        else:
            idx1 = int((high.value - self.f0.value) // self.df.value)
        # crop
        if copy:
            return self[:, idx0:idx1].copy()
        else:
            return self[:, idx0:idx1]

    # -------------------------------------------
    # ufunc modifier

    def _wrap_function(self, function, *args, **kwargs):
        out = super(Spectrogram, self)._wrap_function(
            function, *args, **kwargs)
        # requested frequency axis, return a Spectrum
        if out.ndim == 1 and out.x0.unit == self.y0.unit:
            return Spectrum(out.value, name=out.name, unit=out.unit,
                            epoch=out.epoch, channel=out.channel,
                            f0=out.x0.value, df=out.dx.value)
        # requested time axis, return a TimeSeries
        elif out.ndim == 1:
            return TimeSeries(out.value, name=out.name, unit=out.unit,
                              epoch=out.epoch, channel=out.channel, dx=out.dx)
        # otherwise return whatever we got back from super (Quantity)
        return out
    _wrap_function.__doc__ = Array2D._wrap_function.__doc__


class SpectrogramList(TimeSeriesList):
    """Fancy list representing a list of `Spectrogram`

    The `SpectrogramList` provides an easy way to collect and organise
    `Spectrogram` for a single `Channel` over multiple segments.

    Parameters
    ----------
    *items
        any number of `Spectrogram` series

    Returns
    -------
    list
        a new `SpectrogramList`

    Raises
    ------
    TypeError
        if any elements are not of type `Spectrogram`
    """
    EntryClass = Spectrogram



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from an Omega-format ASCII file.
"""

import os.path
import re
import warnings

from astropy.io.registry import (register_reader, register_identifier)

from glue.lal import CacheEntry

from .ascii import return_reassign_ids
from ..lsctables import SnglBurstTable
from ...io.cache import (file_list, read_cache)
from ...io.utils import (gopen, GzipFile)
from ... import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

SNGL_BURST_COLUMNS = SnglBurstTable.validcolumns.keys()
CWB_ASCII_SNGL_BURST_COLUMN_MAP = {
    'central frequency': 'central_freq',
    'low frequency': 'flow',
    'high frequency': 'fhigh',
    'effective correlated amplitude rho': 'amplitude',
    'likelhood': 'confidence',
    'sSNR': 'snr',
    'time': 'time',
    'time shift': 'time_lag',
}
RE_CWB_ASCII_COLNAME = re.compile('\A#\s+(?P<index>\d+) - (?P<colname>(.*))\Z')
RE_SNGL_COLNAME = re.compile('(?P<name>\w+) for (?P<ifo>[A-Z][0-9]) detector')


def get_column_list(filename, ifo, comments='#'):
    """Read the column list from a cWB ASCII file

    Parameters
    ----------
    filename : `str`
        path of file to read
    ifo : `str`
        prefix of IFO for which to find columns
    comments : `str`, optional, default: '#'
        comment character used for this file

    Returns
    -------
    usecols : `list` of `int`
        the list of column positions to pass to :meth:`numpy.loadtxt`
    columns : `list` of `str`
        the list of column names corresponding to the `usecols` list

    Notes
    -----
    The cWB 'EVENTS.txt' files give a header of (number, name) column
    definitions, however the numbers don't actually match up to the
    actual position of the relevant column when counting from zero, so the
    numbers given in the file are ignored.
    """
    columns = []
    plus2 = 0
    foundifo = False
    with gopen(filename, 'rb') as f:
        while True:
            line = f.readline()
            # stop when the header finishes
            if not line.startswith(comments):
                break
            # work out whether there are + - columns (that aren't counted)
            elif line.startswith('%s -/+' % comments):
                plus2 = 2
                continue
            # parse line, or move on
            try:
                match = RE_CWB_ASCII_COLNAME.match(line.rstrip()).groupdict()
            except AttributeError:
                continue
            colname = match['colname']
            # try and match as a ifo-specific parameter
            try:
                sngl = RE_SNGL_COLNAME.match(colname).groupdict()
            except AttributeError:
                pass
            else:
                colname = sngl['name']
                if sngl['ifo'].lower() != ifo.lower():
                    columns.append(None)
                    continue
                else:
                    foundifo = True
            # otherwise try the custom mapping, then just the sngl_burst name
            try:
                columns.append(CWB_ASCII_SNGL_BURST_COLUMN_MAP[colname])
            except KeyError:
                if colname in SNGL_BURST_COLUMNS:
                    columns.append(colname)
                else:
                    columns.append(None)
    if not foundifo:
        warnings.warn("No %s-specific columns were found in %s"
                      % (ifo, filename), stacklevel=2)
    usecols, names = zip(*[(i+plus2, c) for (i, c) in enumerate(columns)
                           if c is not None])
    return usecols, names


def sngl_burst_table_from_cwb_ascii(f, columns=None, ifo=None, filt=None,
                                    usecols=None, nproc=1, **loadtxtkwargs):
    """Read a `SnglBurstTable` from a cWB-format ASCII file

    Parameters
    ----------
    f : `file`, `str`, `~glue.lal.CacheEntry`, `list`, `~glue.lal.Cache`
        object representing one or more files. One of

        - an open `file`
        - a `str` pointing to a file path on disk
        - a formatted `~glue.lal.CacheEntry` representing one file
        - a `list` of `str` file paths
        - a formatted `~glue.lal.Cache` representing many files

    columns : `list`, optional
        list of column name strings to read, default all.

    ifo : `str`, required
        prefix of IFO to read, required for 'cwb-ascii' format
        (but not for others)

    usecols : `list` of `int`
        the list of column indices to read (absolute, zero-indexed)

    **loadtxtkwargs
        other keyword arguments are passed to `numpy.loadtxt`

    Returns
    -------
    table : `SnglBurstTable`
        a new `SnglBurstTable` filled with data read from the file(s)

    Raises
    ------
    ValueError
        if `ifo` is not given, OR

        if `columns` is given (and not `usecols`) and that column is not
        detected in the file, OR

        if `usecols` is given (and not `columns`) and that column isn't
        parseable from the given file (no `sngl_burst` equivalent)
    """
    # allow multiprocessing
    if nproc != 1:
        return read_cache(f, SnglBurstTable, nproc, return_reassign_ids,
                          columns=columns, usecols=usecols, ifo=ifo,
                          filt=filt, format='cwb-ascii', **loadtxtkwargs)

    comments = loadtxtkwargs.pop('comments', '#')
    files = file_list(f)
    if ifo is None:
        raise ValueError("ifo keyword argument must be given to read cWB "
                         "events from ASCII")
    # parse columns from file header
    allusecols, allcolumns = get_column_list(files[0], ifo, comments=comments)
    # work out which columns the user asked for
    if columns and not usecols:
        given = list(columns)
        columns = []
        usecols = []
        for c in given:
            if c.lower().startswith('peak_time') and 'time' in columns:
                continue
            elif c.lower().startswith('peak_time'):
                c = 'time'
            try:
                index = allcolumns.index(c)
            except ValueError as e:
                e.args = ('Column %r not found in %s for %s'
                          % (c, files[0], ifo),)
                raise
            else:
                columns.append(c)
                usecols.append(allusecols[index])
    elif usecols and not columns:
        columns = []
        for coln in usecols:
            try:
                index = allusecols.index(coln)
            except ValueError as e:
                e.args = ('Column %d not parseable in %s for %s\n'
                          'Note: when reading cWB ASCII files, usecols is the '
                          'list of absolute zero-indexed column positions, not'
                          ' necessarily what is given in the cWB file header'
                          % (coln, files[0], ifo),)
                raise
            else:
                columns.append(allcolumns[index])
    # or use all columns
    elif not columns and not usecols:
        columns = allcolumns
        usecols = allusecols
    # read data
    out = SnglBurstTable.read(files, columns, usecols=usecols, format='ascii',
                              comments=comments, filt=filt, **loadtxtkwargs)
    # append search information and fix SNR
    out.appendColumn('ifo')
    out.appendColumn('search')
    for t in out:
        t.ifo = ifo
        t.search = 'cwb'
        if 'snr' in columns:
            t.snr **= 1/2.
    return out


def sngl_burst_from_cwb(f, *args, **kwargs):
    """Read a `SnglBurstTable` from a cWB file either ROOT of EVENTS.TXT ASCII
    """
    files = file_list(f)
    extensions = list(set(os.path.splitext(fp)[1] for fp in files))
    if len(extensions) == 1 and extensions[0].lower() == '.root':
        raise NotImplementedError("Reading cWB from ROOT files has not been "
                                  "implemented yet")
    elif len(extensions) <= 1:
        return sngl_burst_table_from_cwb_ascii(f, *args, **kwargs)
    raise ValueError("Cannot determine correct cWB reader for multiple file "
                     "extensions: %s" % ", ".join(extensions))


def identify_cwb_ascii(origin, path, fileobj, *args, **kwargs):
    """Automatically identify a fileobj as 'cwb-ascii' format

    Notes
    -----
    This method won't actually let you automatically identify and
    read cwb-ascii files, because the basic 'ascii' format will also
    identify as valid for 'EVENTS.txt' files.

    This function means that when someone runs
    SnglBurstTable.read('EVENTS.txt') an exception will be raised with a
    a sensible error message prompting them to manually specify
    `format='cwb-ascii'`.
    """
    # get file path and filobj
    if isinstance(fileobj, (file, GzipFile)):
        fp = fileobj.name
        pos = fileobj.tell()
        close = False
    elif isinstance(fileobj, CacheEntry):
        fp = fileobj.path
        fileobj = gopen(fp, 'rb')
        close = True
    else:
        fp = fileobj
        fileobj = gopen(fp, 'rb')
        close = True
    # identify by name
    if not os.path.basename(fp) in ['EVENTS.txt', 'EVENTS.txt.gz']:
        return False
    # verify contents
    pos = fileobj.tell()
    try:
        fileobj.seek(0)
        if fileobj.readline().startswith('# correlation threshold'):
            return True
        else:
            return False
    finally:
        fileobj.seek(pos)
        if close:
            fileobj.close()


register_identifier('cwb-ascii', SnglBurstTable, identify_cwb_ascii)
register_reader('cwb-ascii', SnglBurstTable, sngl_burst_table_from_cwb_ascii)
register_reader('cwb', SnglBurstTable, sngl_burst_from_cwb)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Drop in of :mod:`glue.ligolw.lsctables` to annotate Table classes.
"""

from glue.ligolw.lsctables import *

from .. import version
from ..io import reader
from ..time import to_gps

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version
__all__ = TableByName.keys()


def _plot_factory():
    def plot(self, *args, **kwargs):
        """Generate an `EventTablePlot` of this `Table`.

        Parameters
        ----------
        x : `str`
            name of column defining centre point on the X-axis
        y : `str`
            name of column defining centre point on the Y-axis
        width : `str`, optional
            name of column defining width of tile
        height : `str`, optional
            name of column defining height of tile

            .. note::

               The ``width`` and ``height`` positional arguments should
               either both not given, in which case a scatter plot will
               be drawn, or both given, in which case a collections of
               rectangles will be drawn.

        color : `str`, optional, default:`None`
            name of column by which to color markers
        **kwargs
            any other arguments applicable to the `Plot` constructor, and
            the `Table` plotter.

        Returns
        -------
        plot : :class:`~gwpy.plotter.EventTablePlot`
            new plot for displaying tabular data.

        See Also
        --------
        gwpy.plotter.EventTablePlot : for more details.
        """
        from gwpy.plotter import EventTablePlot
        return EventTablePlot(self, *args, **kwargs)

    def hist(self, column, **kwargs):
        """Generate a `HistogramPlot` of this `Table`.

        Parameters
        ----------
        column : `str`
            name of the column over which to histogram data
        **kwargs
            any other arguments applicable to the `HistogramPlot`

        Returns
        -------
        plot : :class:`~gwpy.plotter.HistogramPlot`
            new plot displaying a histogram of this `Table`.
        """
        from gwpy.plotter import HistogramPlot
        return HistogramPlot(self, column, **kwargs)

    return (plot, hist)


def _fetch_factory(table):
    def fetch(cls, channel, etg, start, end, verbose=False, **kwargs):
        """Find and read events into a `{0}`.

        Event XML files are searched for only on the LIGO Data Grid
        using the conventions set out in LIGO-T1300468.

        .. warning::

            This method will not work on machines outside of the LIGO Lab
            computing centres at Caltech, LHO, and LHO.

        Parameters
        ----------
        channel : `str`
            the name of the data channel to search for
        etg : `str`
            the name of the event trigger generator (ETG)
        start : `float`, `~gwpy.time.LIGOTimeGPS`
            the GPS start time of the search
        end : `float`, `~gwpy.time.LIGOTimeGPS`
            the GPS end time of the search
        verbose : `bool`, optional
            print verbose output, default: `False`
        **kwargs
            other keyword arguments to pass to :meth:`{0}.read`

        Returns
        -------
        table : :class:`{0}`
            a new `{0}` containing triggers read from the standard
            paths

        Raises
        ------
        ValueError
            if no channel-level directory is found for the given channel,
            indicating that nothing has ever been processed for this channel.

        See also
        --------
        {0}.read :
            for documentation of the available keyword arguments
        """
        from gwpy.segments import Segment
        from .io.trigfind import find_trigger_urls
        # check times
        start = to_gps(start)
        end = to_gps(end)
        # find files
        cache = find_trigger_urls(channel, etg, start, end, verbose=verbose)
        # construct filter
        infilt = kwargs.pop('filt', None)
        segment = Segment(float(start), float(end))
        if infilt is None:
            def filt(row):
                return float(row.get_peak()) in segment
        else:
            def filt(row):
                return infilt(row) and float(row.get_peak()) in segment
        # read and return
        return cls.read(cache, format='ligolw', filt=filt, **kwargs)
    fetch.__doc__ = fetch.__doc__.format(table.__name__)

    return classmethod(fetch)

# annotate lsctables with new methods
for table in TableByName.itervalues():
    # define the read classmethod with docstring
    table.read = classmethod(reader(doc="""
        Read data into a `{0}`.

        Parameters
        ----------
        f : `file`, `str`, `~glue.lal.CacheEntry`, `list`, `~glue.lal.Cache`
            object representing one or more files. One of

                - an open `file`
                - a `str` pointing to a file path on disk
                - a formatted `~glue.lal.CacheEntry` representing one file
                - a `list` of `str` file paths
                - a formatted `~glue.lal.Cache` representing many files

        columns : `list`, optional
            list of column name strings to read, default all.

        ifo : `str`, optional
            prefix of IFO to read

            .. warning::

               the ``ifo`` keyword argument is only applicable (but is
               required) when reading single-interferometer data from
               a multi-interferometer file

        filt : `function`, optional
            function by which to `filter` events. The callable must
            accept as input a row of the table event and return
            `True`/`False`.

        nproc : `int`, optional, default: ``1``
            number of parallel processes with which to distribute file I/O,
            default: serial process.

            .. warning::

               The ``nproc`` keyword argument is only applicable when
               reading a `list` (or `~glue.lal.Cache`) of files.

        contenthandler : `~glue.ligolw.ligolw.LIGOLWContentHandler`
            SAX content handler for parsing ``LIGO_LW`` documents.

            .. warning::

               The ``contenthandler`` keyword argument is only applicable
               when reading from ``LIGO_LW`` documents.

        **loadtxtkwargs
            when reading from ASCII, all other keyword arguments are passed
            directly to `numpy.loadtxt`

        Returns
        -------
        table : :class:`{0}`
            `{0}` of data with given columns filled

        Notes
        -----
        """.format(table.__name__)))
    if ('start_time' in table.validcolumns or
            'peak_time' in table.validcolumns or
            'end_time' in table.validcolumns):
        table.plot, table.hist = _plot_factory()

    if table == SnglBurstTable:
        table.fetch = _fetch_factory(table)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""A `Segment` is a interval of time marked by a GPS [start, stop)
semi-open interval. These typically represent periods when a
gravitational-wave laser interferometer was operating in a specific
configuration.
"""

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Kipp Cannon <kipp.cannon@ligo.org>"

from glue.segments import (segment as _Segment,
                           segmentlist as _SegmentList,
                           segmentlistdict as _SegmentListDict)

from ..io import (reader, writer)


def _update_docstring(doc):
    """Update the docstring of the given class from glue to match
    the new naming for GWpy matching PEP-8.

    Parameters
    ----------
    doc : `str`
        input docstring to modify

    Returns
    -------
    newdoc : `str`
        the updated docstring with all references to the
        glue segments.segment* classes modified for GWpy with sphinx
        scope
    """
    doc = doc.replace('segmentlistdicts',
                      '`SegmentListDicts <SegmentListDict>`')
    doc = doc.replace('segmentlistdict', '`SegmentListDict`')
    doc = doc.replace('Segmentlists', '`SegmentLists <SegmentList>`')
    doc = doc.replace('segmentlist', '`SegmentList`')
    doc = doc.replace('segments', '`Segments <Segment>`')
    doc = doc.replace('segment', '`Segment`')
    return doc


class Segment(_Segment):
    __doc__ = _update_docstring(_Segment.__doc__)

    @property
    def start(self):
        """The GPS start time of this segment
        """
        return self[0]

    @property
    def end(self):
        """The GPS end time of this segment
        """
        return self[1]

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self[0], self[1])

    def __str__(self):
        return "[%s ... %s)" % (self[0], self[1])


class SegmentList(_SegmentList):
    __doc__ = _update_docstring(_SegmentList.__doc__)

    def __repr__(self):
        return "<SegmentList([%s])>" % "\n              ".join(map(repr, self))

    def coalesce(self):
        self = super(SegmentList, self).coalesce()
        for i, s in enumerate(self):
            self[i] = Segment(s[0], s[1])
        return self
    coalesce.__doc__ = _SegmentList.coalesce.__doc__

    def __str__(self):
        return "[%s]" % "\n ".join(map(str, self))

    read = classmethod(reader(doc="""
    Read segments from file into a `SegmentList`.

    Parameters
    ----------
    filename : `str`
        path of file to read

    format : `str`, optional
        source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    flag : `str`, optional, default: read all segments
        name of flag to read from file.

    gpstype : `type`, optional, default: `float`
        datatype to force for segment GPS times

    strict : `bool`, optional, default: `True`
        require segment start and stop times match printed duration,
        only valid for ``format='segwizard'``.

    Returns
    -------
    segmentlist : `SegmentList`
        `SegmentList` active and known segments read from file.

    Notes
    -----
    When reading with ``format='segwizard'`` the
    :attr:`~DataQualityFlag.known` `SegmentList` will simply represent
    the extent of the :attr:`~DataQualityFlag.active` `SegmentList`.
    """))
    write = writer()


class SegmentListDict(_SegmentListDict):
    __doc__ = _update_docstring(_SegmentListDict.__doc__)

    read = classmethod(reader(doc="""
    Read segments from file into a `SegmentListDict`.

    Parameters
    ----------
    filename : `str`
        path of file to read

    format : `str`, optional
        source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    flags : `list`, optional, default: read all flags
        name of flag to read from file.

    gpstype : `type`, optional, default: `float`
        datatype to force for segment GPS times

    strict : `bool`, optional, default: `True`
        require segment start and stop times match printed duration,
        only valid for ``format='segwizard'``.

    Returns
    -------
    dqflag : `DataQualityFlag`
        formatted `DataQualityFlag` containing the active and known
        segments read from file.

    Notes
    -----
    When reading with ``format='segwizard'`` the
    :attr:`~DataQualityFlag.known` `SegmentList` will simply represent
    the extent of the :attr:`~DataQualityFlag.active` `SegmentList`.
    """))
    write = writer()

del _Segment
del _SegmentList
del _SegmentListDict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

""" Coherence plots
"""
from cliproduct import CliProduct


class Coherence(CliProduct):
    """Product class for coherence line plots"""

    min_timeseries = 2
    plot = 0    # this will be a matplotlib plot after derived class makes it

    def get_action(self):
        """Return the string used as "action" on command line."""
        return 'coherence'

    def init_cli(self, parser):
        """Set up the argument list for this product"""
        self.arg_chan2(parser)
        self.arg_freq(parser)
        self.arg_ax_xlf(parser)
        self.arg_ax_liny(parser)
        self.arg_plot(parser)
        self.xaxis_is_freq = True
        return

    def get_ylabel(self, args):
        """Text for y-axis label"""
        return 'Coherence'

    def get_title(self):
        """Start of default super title, first channel is appended to it"""
        return 'Coherence with: '

    def get_min_datasets(self):
        """Coherence requires 2 datasets for the calculation"""
        return 2

    def get_xlabel(self):
        return 'Frequency (Hz)'

    def get_sup_title(self):
        """Override if default lacks critical info"""
        return self.get_title() + self.ref_chan_name

    def gen_plot(self, arg_list):
        """Generate the coherence plot from all time series"""
        import numpy

        self.is_freq_plot = True
        fftlen = 1
        if arg_list.secpfft:
            fftlen = float(arg_list.secpfft)
        ovlap_frac = 0.5
        if arg_list.overlap:
            ovlap_frac = float(arg_list.overlap)
        self.secpfft = fftlen
        self.overlap = ovlap_frac
        maxfs = 0

        if arg_list.ref:
            ref_name = arg_list.ref
        else:
            ref_name = self.timeseries[0].channel.name
        self.ref_chan_name = ref_name

        self.log(3, 'Reference channel: ' + ref_name)
        # we don't want to compare the reference channel to itself
        # at a different time
        next_ts = -1
        for idx in range(0, len(self.timeseries)):
            legend_text = self.timeseries[idx].channel.name
            if legend_text != ref_name and self.timeseries[idx].min() != \
                    self.timeseries[idx].value.max():
                next_ts = idx
                break
        if next_ts == -1:
            raise ValueError('No appropriate channels for '
                             'Coherence calculation')

        cohs = []
        for time_group in self.time_groups:
            ref_idx = time_group[0]
            if len(time_group) >= 2:
                # find the reference channel in this group
                for idx in range(0, len(time_group)):
                    idxp = time_group[idx]
                    if self.timeseries[idxp].channel.name == ref_name:
                        ref_idx = idxp

                maxfs = max(maxfs, self.timeseries[ref_idx].sample_rate)
                if numpy.min(self.timeseries[ref_idx]) == \
                        numpy.max(self.timeseries[ref_idx]):
                    print 'Channel %s at %d has min=max,it cannot be used ' \
                          'as a reference channel' \
                          % self.timeseries[ref_idx].channel.name, \
                        self.timeseries[ref_idx].epoch.gps
                else:
                    for idxp in range(0, len(time_group)):
                        next_ts = time_group[idxp]
                        if next_ts != ref_idx:
                            if numpy.min(self.timeseries[next_ts]) == \
                                    numpy.max(self.timeseries[next_ts]):
                                print 'Channel %s at %d has min=max, ' \
                                      'coherence with this channel will not ' \
                                      'be calculated' \
                                      % self.timeseries[next_ts].channel.name,\
                                    self.timeseries[next_ts].epoch.gps
                            else:
                                maxfs = max(maxfs,
                                            self.timeseries[next_ts].
                                            sample_rate)
                                # calculate and plot the first pair,
                                # note the first channel is the reference chan
                                snd_ts = self.timeseries[next_ts]
                                coh = self.timeseries[ref_idx].\
                                    coherence(snd_ts, fftlength=fftlen,
                                              overlap=ovlap_frac*fftlen)

                                legend_text = self.timeseries[next_ts].\
                                    channel.name
                                if len(self.start_list) > 1:
                                    legend_text += ", %s" % \
                                                   snd_ts.times.epoch.gps
                                coh.name = legend_text

                                # coh2 = 1 / (1-coh) : how to do alt scaler

                                if not cohs:
                                    self.plot = coh.plot()
                                else:
                                    self.plot.add_spectrum(coh)

                                cohs.append(coh)

        if not cohs:
            raise ValueError('No coherence was calculated due to data'
                             ' problems (avaiability or constant values)')
        # if the specified frequency limits adjust our ymin and ymax values
        # at this point self.ymin and self.ymax represent the full spectra
        mymin = cohs[0].value.min()
        mymax = cohs[0].value.max()
        myfmin = 1/self.secpfft * cohs[0].frequencies.unit
        myfmax = maxfs/2
        if arg_list.fmin:
            myfmin = float(arg_list.fmin) * cohs[0].frequencies.unit
        if arg_list.fmax:
            myfmax = float(arg_list.fmax) * cohs[0].frequencies.unit

        for idx in range(0, len(cohs)):
            t = numpy.where(cohs[idx].frequencies >= myfmin)
            if t[0].size:
                strt = t[0][0]
                t = numpy.where(cohs[idx].frequencies >= myfmax)
                if t[0].size:
                    stop = t[0][t[0].size-1]
                else:
                    stop = cohs[idx].frequencies.size - 1
                mymin = min(mymin, numpy.min(cohs[idx].value[strt:stop]))
                mymax = max(mymax, numpy.max(cohs[idx].value[strt:stop]))

        self.ymin = mymin
        self.ymax = mymax
        self.fmin = max(myfmin.value, 1/self.secpfft)
        self.fmax = min(myfmax.value, maxfs.value/2)

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the GWpy package
"""

from __future__ import print_function

import sys
if sys.version < '2.6':
    raise ImportError("Python versions older than 2.6 are not supported.")

import glob
import hashlib
import os.path
import subprocess

try:
    import setuptools
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
finally:
    from setuptools import (setup, find_packages)
    from setuptools.command import (build_py, egg_info)

from distutils import log
from distutils.dist import Distribution
from distutils.cmd import Command
from distutils.command.clean import (clean, log, remove_tree)

# test for OrderedDict
extra_install_requires = []
try:
    from collections import OrderedDict
except ImportError:
    extra_install_requires.append('ordereddict>=1.1')

# importlib required for cli programs
try:
    from importlib import import_module
except ImportError:
    extra_install_requires.append('importlib>=1.0.3')

# test for unittest2
extra_tests_require = []
if sys.version < '2.7':
    extra_tests_require.append('unittest2')

# import sphinx commands
try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    cmdclass = {}
else:
    cmdclass = {'build_sphinx': BuildDoc}

# set basic metadata
PACKAGENAME = 'gwpy'
AUTHOR = 'Duncan Macleod'
AUTHOR_EMAIL = 'duncan.macleod@ligo.org'
LICENSE = 'GPLv3'

VERSION_PY = os.path.join(PACKAGENAME, 'version.py')


# -----------------------------------------------------------------------------
# Clean up, including Sphinx, and setup_requires eggs

class GWpyClean(clean):
    def run(self):
        if self.all:
            # remove dist
            if os.path.exists('dist'):
                remove_tree('dist')
            else:
                log.warn("'dist' does not exist -- can't clean it")
            # remove docs
            sphinx_dir = os.path.join(self.build_base, 'sphinx')
            if os.path.exists(sphinx_dir):
                remove_tree(sphinx_dir, dry_run=self.dry_run)
            else:
                log.warn("%r does not exist -- can't clean it", sphinx_dir)
            # remove version.py
            for vpy in [VERSION_PY, VERSION_PY + 'c']:
                if os.path.exists(vpy) and not self.dry_run:
                    log.info('removing %r' % vpy)
                    os.unlink(vpy)
                elif not os.path.exists(vpy):
                    log.warn("%r does not exist -- can't clean it", vpy)
            # remove setup eggs
            for egg in glob.glob('*.egg'):
                if os.path.isdir(egg):
                    remove_tree(egg, dry_run=self.dry_run)
                else:
                    log.info('removing %r' % egg)
                    os.unlink(egg)
            # remove Portfile
            portfile = 'Portfile'
            if os.path.exists(portfile) and not self.dry_run:
                log.info('removing %r' % portfile)
                os.unlink(portfile)
        clean.run(self)

cmdclass['clean'] = GWpyClean


# -----------------------------------------------------------------------------
# Custom builders to write version.py

class GitVersionMixin(object):
    """Mixin class to add methods to generate version information from git.
    """
    def write_version_py(self, pyfile):
        """Generate target file with versioning information from git VCS
        """
        log.info("generating %s" % pyfile)
        import gwpy._version_helper as vcs
        gitstatus = vcs.GitStatus()
        try:
            with open(pyfile, 'w') as fobj:
                gitstatus.write(fobj, author=AUTHOR, email=AUTHOR_EMAIL)
        except:
            if os.path.exists(pyfile):
                os.unlink(pyfile)
            raise
        return gitstatus

    def update_metadata(self):
        """Import package base and update distribution metadata
        """
        import gwpy
        self.distribution.metadata.version = gwpy.__version__
        desc, longdesc = gwpy.__doc__.split('\n', 1)
        self.distribution.metadata.description = desc
        self.distribution.metadata.long_description = longdesc.strip('\n')


class GWpyBuildPy(build_py.build_py, GitVersionMixin):
    """Custom build_py command to deal with version generation
    """
    def __init__(self, *args, **kwargs):
        build_py.build_py.__init__(self, *args, **kwargs)

    def run(self):
        try:
            self.write_version_py(VERSION_PY)
        except ImportError:
            raise
        except:
            if not os.path.isfile(VERSION_PY):
                raise
        self.update_metadata()
        build_py.build_py.run(self)

cmdclass['build_py'] = GWpyBuildPy


class GWpyEggInfo(egg_info.egg_info, GitVersionMixin):
    """Custom egg_info command to deal with version generation
    """
    def finalize_options(self):
        try:
            self.write_version_py(VERSION_PY)
        except ImportError:
            raise
        except:
            if not os.path.isfile(VERSION_PY):
                raise
        if not self.distribution.metadata.version:
            self.update_metadata()
        egg_info.egg_info.finalize_options(self)

cmdclass['egg_info'] = GWpyEggInfo


# -----------------------------------------------------------------------------
# Build Portfile

class BuildPortfile(Command, GitVersionMixin):
    """Generate a Macports Portfile for this project from the current build
    """
    description = 'Generate Macports Portfile'
    user_options = [
        ('version=', None, 'the X.Y.Z package version'),
        ('portfile=', None, 'target output file, default: \'Portfile\''),
        ('template=', None,
         'Portfile template, default: \'Portfile.template\''),
    ]

    def initialize_options(self):
        self.version = None
        self.portfile = 'Portfile'
        self.template = 'Portfile.template'
        self._template = None

    def finalize_options(self):
        from jinja2 import Template
        with open(self.template, 'r') as t:
            self._template = Template(t.read())

    def run(self):
        # get version from distribution
        if self.version is None:
            try:
                self.update_metadata()
            except ImportError:
                self.run_command('sdist')
                self.update_metadata()
        # find dist file
        dist = os.path.join(
            'dist',
            '%s-%s.tar.gz' % (self.distribution.get_name(),
                              self.distribution.get_version()))
        # run sdist if needed
        if not os.path.isfile(dist):
            self.run_command('sdist')
            self.update_metadata()
        # get checksum digests
        log.info('reading distribution tarball %r' % dist)
        with open(dist, 'rb') as fobj:
            data = fobj.read()
        log.info('recovered digests:')
        digest = dict()
        digest['rmd160'] = self._get_rmd160(dist)
        for algo in [1, 256]:
            digest['sha%d' % algo] = self._get_sha(data, algo)
        for key, val in digest.iteritems():
            log.info('    %s: %s' % (key, val))
        # write finished portfile to file
        with open(self.portfile, 'w') as fport:
            fport.write(self._template.render(
                version=self.distribution.get_version(), **digest))
        log.info('portfile written to %r' % self.portfile)

    @staticmethod
    def _get_sha(data, algorithm=256):
        hash_ = getattr(hashlib, 'sha%d' % algorithm)
        return hash_(data).hexdigest()

    @staticmethod
    def _get_rmd160(filename):
        p = subprocess.Popen(['openssl', 'rmd160', filename],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if p.returncode != 0:
            raise subprocess.CalledProcessError(err)
        else:
            return out.splitlines()[0].rsplit(' ', 1)[-1]

cmdclass['port'] = BuildPortfile


# -----------------------------------------------------------------------------
# don't use setup_requires if just checking for information

# (credit: matplotlib/setup.py)
setup_requires = []
if '--help' not in sys.argv and '--help-commands' not in sys.argv:
    dist_ = Distribution({'cmdclass': cmdclass})
    dist_.parse_config_files()
    dist_.parse_command_line()
    if not (any('--' + opt in sys.argv for opt in
            Distribution.display_option_names + ['help']) or
            dist_.commands == ['clean']):
        setup_requires = ['tornado', 'numpy >= 1.7', 'jinja2', 'gitpython']

# -----------------------------------------------------------------------------
# Find files

# Use the find_packages tool to locate all packages and modules
packagenames = find_packages()

# glob for all scripts
if os.path.isdir('bin'):
    scripts = glob.glob(os.path.join('bin', '*'))
else:
    scripts = []

# -----------------------------------------------------------------------------
# run setup

setup(name=PACKAGENAME,
      provides=[PACKAGENAME],
      version=None,
      description=None,
      long_description=None,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url='https://gwpy.github.io/',
      packages=packagenames,
      include_package_data=True,
      cmdclass=cmdclass,
      scripts=scripts,
      setup_requires=setup_requires,
      requires=[
          'glue',
          'dateutil',
          'numpy',
          'scipy',
          'matplotlib',
          'astropy',
          'six',
      ],
      install_requires=[
          'python-dateutil',
          'numpy >= 1.7',
          'scipy >= 0.16.0',
          'matplotlib >= 1.3.0',
          'astropy >= 1.0',
          'six >= 1.5',
      ] + extra_install_requires,
      tests_require=[
      ] + extra_tests_require,
      extras_require={
          'nds': ['nds2-client'],
          'gwf': ['ldas-tools'],
          'doc': ['sphinx', 'numpydoc', 'sphinx-bootstrap-theme',
                  'sphinxcontrib-doxylink', 'sphinxcontrib-epydoc',
                  'sphinxcontrib-programoutput'],
          'hdf5': ['h5py'],
      },
      dependency_links=[
          'https://www.lsc-group.phys.uwm.edu/daswg/download/'
          'software/source/glue-1.48.tar.gz#egg=glue-1.48',
      ],
      test_suite='gwpy.tests',
      use_2to3=False,
      classifiers=[
          'Programming Language :: Python',
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Topic :: Scientific/Engineering :: Physics',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Operating System :: MacOS',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      ],
      )
